/*
 * Copyright 1999-2008 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package edu.rice.cs.mint.comp.com.sun.tools.javac.comp;

import java.util.*;
import java.util.Set;
import edu.rice.cs.mint.comp.javax.lang.model.element.ElementKind;

import edu.rice.cs.mint.comp.com.sun.tools.javac.code.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.jvm.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.tree.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.util.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.util.JCDiagnostic.DiagnosticPosition;
import edu.rice.cs.mint.comp.com.sun.tools.javac.util.List;

import edu.rice.cs.mint.comp.com.sun.tools.javac.jvm.Target;
import edu.rice.cs.mint.comp.com.sun.tools.javac.code.Symbol.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.tree.JCTree.*;
import edu.rice.cs.mint.comp.com.sun.tools.javac.code.Type.*;

import edu.rice.cs.mint.comp.com.sun.source.tree.IdentifierTree;
import edu.rice.cs.mint.comp.com.sun.source.tree.MemberSelectTree;
import edu.rice.cs.mint.comp.com.sun.source.tree.TreeVisitor;
import edu.rice.cs.mint.comp.com.sun.source.util.SimpleTreeVisitor;

import static edu.rice.cs.mint.comp.com.sun.tools.javac.code.Flags.*;
import static edu.rice.cs.mint.comp.com.sun.tools.javac.code.Kinds.*;
import static edu.rice.cs.mint.comp.com.sun.tools.javac.code.TypeTags.*;

/** This is the main context-dependent analysis phase in GJC. It
 *  encompasses name resolution, type checking and constant folding as
 *  subtasks. Some subtasks involve auxiliary classes.
 *  @see Check
 *  @see Resolve
 *  @see ConstFold
 *  @see Infer
 *
 *  <p><b>This is NOT part of any API supported by Sun Microsystems.  If
 *  you write code that depends on this, you do so at your own risk.
 *  This code and its internal interfaces are subject to change or
 *  deletion without notice.</b>
 */
public class Attr extends JCTree.Visitor {
    protected static final Context.Key<Attr> attrKey =
        new Context.Key<Attr>();

    final Names names;
    final Log log;
    final Symtab syms;
    final Resolve rs;
    final Check chk;
    final MemberEnter memberEnter;
    final TreeMaker make;
    final ConstFold cfolder;
    final Enter enter;
    final Target target;
    final Types types;
    final JCDiagnostic.Factory diags;
    final Annotate annotate;
    final Options options;

    public static Attr instance(Context context) {
        Attr instance = context.get(attrKey);
        if (instance == null)
            instance = new Attr(context);
        return instance;
    }

    protected Attr(Context context) {
        context.put(attrKey, this);

        names = Names.instance(context);
        log = Log.instance(context);
        syms = Symtab.instance(context);
        rs = Resolve.instance(context);
        chk = Check.instance(context);
        memberEnter = MemberEnter.instance(context);
        make = TreeMaker.instance(context);
        enter = Enter.instance(context);
        cfolder = ConstFold.instance(context);
        target = Target.instance(context);
        types = Types.instance(context);
        diags = JCDiagnostic.Factory.instance(context);
        annotate = Annotate.instance(context);

        options = Options.instance(context);

        Source source = Source.instance(context);
        allowGenerics = source.allowGenerics();
        allowVarargs = source.allowVarargs();
        allowEnums = source.allowEnums();
        allowBoxing = source.allowBoxing();
        allowCovariantReturns = source.allowCovariantReturns();
        allowAnonOuterThis = source.allowAnonOuterThis();
        relax = (options.get("-retrofit") != null ||
                 options.get("-relax") != null);
        useBeforeDeclarationWarning = options.get("useBeforeDeclarationWarning") != null;
    }

    /** Switch: relax some constraints for retrofit mode.
     */
    boolean relax;

    /** Switch: support generics?
     */
    boolean allowGenerics;

    /** Switch: allow variable-arity methods.
     */
    boolean allowVarargs;

    /** Switch: support enums?
     */
    boolean allowEnums;

    /** Switch: support boxing and unboxing?
     */
    boolean allowBoxing;

    /** Switch: support covariant result types?
     */
    boolean allowCovariantReturns;

    /** Switch: allow references to surrounding object from anonymous
     * objects during constructor call?
     */
    boolean allowAnonOuterThis;

    /**
     * Switch: warn about use of variable before declaration?
     * RFE: 6425594
     */
    boolean useBeforeDeclarationWarning;

    /** Check kind and type of given tree against protokind and prototype.
     *  If check succeeds, store type in tree and return it.
     *  If check fails, store errType in tree and return it.
     *  No checks are performed if the prototype is a method type.
     *  It is not necessary in this case since we know that kind and type
     *  are correct.
     *
     *  @param tree     The tree whose kind and type is checked
     *  @param owntype  The computed type of the tree
     *  @param ownkind  The computed kind of the tree
     *  @param pkind    The expected kind (or: protokind) of the tree
     *  @param pt       The expected type (or: prototype) of the tree
     */
    Type check(JCTree tree, Type owntype, int ownkind, int pkind, Type pt) {
        if (owntype.tag != ERROR && pt.tag != METHOD && pt.tag != FORALL) {
            if ((ownkind & ~pkind) == 0) {
                owntype = chk.checkType(tree.pos(), owntype, pt);
            } else {
                log.error(tree.pos(), "unexpected.type",
                          kindNames(pkind),
                          kindName(ownkind));
                owntype = types.createErrorType(owntype);
            }
        }
        tree.type = owntype;
        return owntype;
    }

    /** Is given blank final variable assignable, i.e. in a scope where it
     *  may be assigned to even though it is final?
     *  @param v      The blank final variable.
     *  @param env    The current environment.
     */
    boolean isAssignableAsBlankFinal(VarSymbol v, Env<AttrContext> env) {
        Symbol owner = env.info.scope.owner;
           // owner refers to the innermost variable, method or
           // initializer block declaration at this point.
        return
            v.owner == owner
            ||
            ((owner.name == names.init ||    // i.e. we are in a constructor
              owner.kind == VAR ||           // i.e. we are in a variable initializer
              (owner.flags() & BLOCK) != 0)  // i.e. we are in an initializer block
             &&
             v.owner == owner.owner
             &&
             ((v.flags() & STATIC) != 0) == Resolve.isStatic(env));
    }

    /** Check that variable can be assigned to.
     *  @param pos    The current source code position.
     *  @param v      The assigned varaible
     *  @param base   If the variable is referred to in a Select, the part
     *                to the left of the `.', null otherwise.
     *  @param env    The current environment.
     */
    void checkAssignable(DiagnosticPosition pos, VarSymbol v, JCTree base, Env<AttrContext> env) {
        if ((v.flags() & FINAL) != 0 &&
            ((v.flags() & HASINIT) != 0
             ||
             !((base == null ||
               (base.getTag() == JCTree.IDENT && TreeInfo.name(base) == names._this)) &&
               isAssignableAsBlankFinal(v, env)))) {
            log.error(pos, "cant.assign.val.to.final.var", v);
        }
    }

    /** Does tree represent a static reference to an identifier?
     *  It is assumed that tree is either a SELECT or an IDENT.
     *  We have to weed out selects from non-type names here.
     *  @param tree    The candidate tree.
     */
    boolean isStaticReference(JCTree tree) {
        if (tree.getTag() == JCTree.SELECT) {
            Symbol lsym = TreeInfo.symbol(((JCFieldAccess) tree).selected);
            if (lsym == null || lsym.kind != TYP) {
                return false;
            }
        }
        return true;
    }

    /** Is this symbol a type?
     */
    static boolean isType(Symbol sym) {
        return sym != null && sym.kind == TYP;
    }

    /** The current `this' symbol.
     *  @param env    The current environment.
     */
    Symbol thisSym(DiagnosticPosition pos, Env<AttrContext> env) {
        return rs.resolveSelf(pos, env, env.enclClass.sym, names._this);
    }

    /** Attribute a parsed identifier.
     * @param tree Parsed identifier name
     * @param topLevel The toplevel to use
     */
    public Symbol attribIdent(JCTree tree, JCCompilationUnit topLevel) {
        Env<AttrContext> localEnv = enter.topLevelEnv(topLevel);
        localEnv.enclClass = make.ClassDef(make.Modifiers(0),
                                           syms.errSymbol.name,
                                           null, null, null, null);
        localEnv.enclClass.sym = syms.errSymbol;
        return tree.accept(identAttributer, localEnv);
    }
    // where
        private TreeVisitor<Symbol,Env<AttrContext>> identAttributer = new IdentAttributer();
        private class IdentAttributer extends SimpleTreeVisitor<Symbol,Env<AttrContext>> {
            @Override
            public Symbol visitMemberSelect(MemberSelectTree node, Env<AttrContext> env) {
                Symbol site = visit(node.getExpression(), env);
                if (site.kind == ERR)
                    return site;
                Name name = (Name)node.getIdentifier();
                if (site.kind == PCK) {
                    env.toplevel.packge = (PackageSymbol)site;
                    return rs.findIdentInPackage(env, (TypeSymbol)site, name, TYP | PCK);
                } else {
                    env.enclClass.sym = (ClassSymbol)site;
                    return rs.findMemberType(env, site.asType(), name, (TypeSymbol)site);
                }
            }

            @Override
            public Symbol visitIdentifier(IdentifierTree node, Env<AttrContext> env) {
                return rs.findIdent(env, (Name)node.getName(), TYP | PCK);
            }
        }

    public Type coerce(Type etype, Type ttype) {
        return cfolder.coerce(etype, ttype);
    }

    public Type attribType(JCTree node, TypeSymbol sym) {
        Env<AttrContext> env = enter.typeEnvs.get(sym);
        Env<AttrContext> localEnv = env.dup(node, env.info.dup());
        return attribTree(node, localEnv, Kinds.TYP, Type.noType);
    }

    public Env<AttrContext> attribExprToTree(JCTree expr, Env<AttrContext> env, JCTree tree) {
        breakTree = tree;
        edu.rice.cs.mint.comp.javax.tools.JavaFileObject prev = log.useSource(null);
        try {
            attribExpr(expr, env);
        } catch (BreakAttr b) {
            return b.env;
        } finally {
            breakTree = null;
            log.useSource(prev);
        }
        return env;
    }

    public Env<AttrContext> attribStatToTree(JCTree stmt, Env<AttrContext> env, JCTree tree) {
        breakTree = tree;
        edu.rice.cs.mint.comp.javax.tools.JavaFileObject prev = log.useSource(null);
        try {
            attribStat(stmt, env);
        } catch (BreakAttr b) {
            return b.env;
        } finally {
            breakTree = null;
            log.useSource(prev);
        }
        return env;
    }

    private JCTree breakTree = null;

    private static class BreakAttr extends RuntimeException {
        static final long serialVersionUID = -6924771130405446405L;
        private Env<AttrContext> env;
        private BreakAttr(Env<AttrContext> env) {
            this.env = env;
        }
    }


/* ************************************************************************
 * Visitor methods
 *************************************************************************/

    /** Visitor argument: the current environment.
     */
    Env<AttrContext> env;

    /** Visitor argument: the currently expected proto-kind.
     */
    int pkind;

    /** Visitor argument: the currently expected proto-type.
     */
    Type pt;

    /** Visitor result: the computed type.
     */
    Type result;

    /** Visitor method: attribute a tree, catching any completion failure
     *  exceptions. Return the tree's type.
     *
     *  @param tree    The tree to be visited.
     *  @param env     The environment visitor argument.
     *  @param pkind   The protokind visitor argument.
     *  @param pt      The prototype visitor argument.
     */
    Type attribTree(JCTree tree, Env<AttrContext> env, int pkind, Type pt) {
        Env<AttrContext> prevEnv = this.env;
        int prevPkind = this.pkind;
        Type prevPt = this.pt;
        try {
            this.env = env;
            this.pkind = pkind;
            this.pt = pt;
            tree.accept(this);
            if (tree == breakTree)
                throw new BreakAttr(env);
            return result;
        } catch (CompletionFailure ex) {
            tree.type = syms.errType;
            return chk.completionError(tree.pos(), ex);
        } finally {
            this.env = prevEnv;
            this.pkind = prevPkind;
            this.pt = prevPt;
        }
    }

    /** Derived visitor method: attribute an expression tree.
     */
    public Type attribExpr(JCTree tree, Env<AttrContext> env, Type pt) {
        return attribTree(tree, env, VAL, pt.tag != ERROR ? pt : Type.noType);
    }

    /** Derived visitor method: attribute an expression tree with
     *  no constraints on the computed type.
     */
    Type attribExpr(JCTree tree, Env<AttrContext> env) {
        return attribTree(tree, env, VAL, Type.noType);
    }

    /** Derived visitor method: attribute a type tree.
     */
    Type attribType(JCTree tree, Env<AttrContext> env) {
        Type result = attribTree(tree, env, TYP, Type.noType);
        return result;
    }

    /** Derived visitor method: attribute a statement or definition tree.
     */
    public Type attribStat(JCTree tree, Env<AttrContext> env) {
        return attribTree(tree, env, NIL, Type.noType);
    }

    /** Attribute a list of expressions, returning a list of types.
     */
    List<Type> attribExprs(List<JCExpression> trees, Env<AttrContext> env, Type pt) {
        ListBuffer<Type> ts = new ListBuffer<Type>();
        for (List<JCExpression> l = trees; l.nonEmpty(); l = l.tail)
            ts.append(attribExpr(l.head, env, pt));
        return ts.toList();
    }

    /** Attribute a list of statements, returning nothing.
     */
    <T extends JCTree> void attribStats(List<T> trees, Env<AttrContext> env) {
        for (List<T> l = trees; l.nonEmpty(); l = l.tail)
            attribStat(l.head, env);
    }

    /** Attribute the arguments in a method call, returning a list of types.
     */
    List<Type> attribArgs(List<JCExpression> trees, Env<AttrContext> env) {
        ListBuffer<Type> argtypes = new ListBuffer<Type>();
        for (List<JCExpression> l = trees; l.nonEmpty(); l = l.tail)
            argtypes.append(chk.checkNonVoid(
                l.head.pos(), types.upperBound(attribTree(l.head, env, VAL, Infer.anyPoly))));
        return argtypes.toList();
    }

    /** Attribute a type argument list, returning a list of types.
     */
    List<Type> attribTypes(List<JCExpression> trees, Env<AttrContext> env) {
        ListBuffer<Type> argtypes = new ListBuffer<Type>();
        for (List<JCExpression> l = trees; l.nonEmpty(); l = l.tail)
            argtypes.append(chk.checkRefType(l.head.pos(), attribType(l.head, env)));
        return argtypes.toList();
    }


    /**
     * Attribute type variables (of generic classes or methods).
     * Compound types are attributed later in attribBounds.
     * @param typarams the type variables to enter
     * @param env      the current environment
     */
    void attribTypeVariables(List<JCTypeParameter> typarams, Env<AttrContext> env) {
        for (JCTypeParameter tvar : typarams) {
            TypeVar a = (TypeVar)tvar.type;
            a.tsym.flags_field |= UNATTRIBUTED;
            a.bound = Type.noType;
            if (!tvar.bounds.isEmpty()) {
                List<Type> bounds = List.of(attribType(tvar.bounds.head, env));
                for (JCExpression bound : tvar.bounds.tail)
                    bounds = bounds.prepend(attribType(bound, env));
                types.setBounds(a, bounds.reverse());
            } else {
                // if no bounds are given, assume a single bound of
                // java.lang.Object.
                types.setBounds(a, List.of(syms.objectType));
            }
            a.tsym.flags_field &= ~UNATTRIBUTED;
        }
        for (JCTypeParameter tvar : typarams)
            chk.checkNonCyclic(tvar.pos(), (TypeVar)tvar.type);
        attribStats(typarams, env);
    }

    void attribBounds(List<JCTypeParameter> typarams) {
        for (JCTypeParameter typaram : typarams) {
            Type bound = typaram.type.getUpperBound();
            if (bound != null && bound.tsym instanceof ClassSymbol) {
                ClassSymbol c = (ClassSymbol)bound.tsym;
                if ((c.flags_field & COMPOUND) != 0) {
                    assert (c.flags_field & UNATTRIBUTED) != 0 : c;
                    attribClass(typaram.pos(), c);
                }
            }
        }
    }

    /**
     * Attribute the type references in a list of annotations.
     */
    void attribAnnotationTypes(List<JCAnnotation> annotations,
                               Env<AttrContext> env) {
        for (List<JCAnnotation> al = annotations; al.nonEmpty(); al = al.tail) {
            JCAnnotation a = al.head;
            attribType(a.annotationType, env);
        }
    }

    /** Attribute type reference in an `extends' or `implements' clause.
     *
     *  @param tree              The tree making up the type reference.
     *  @param env               The environment current at the reference.
     *  @param classExpected     true if only a class is expected here.
     *  @param interfaceExpected true if only an interface is expected here.
     */
    Type attribBase(JCTree tree,
                    Env<AttrContext> env,
                    boolean classExpected,
                    boolean interfaceExpected,
                    boolean checkExtensible) {
        Type t = attribType(tree, env);
        return checkBase(t, tree, env, classExpected, interfaceExpected, checkExtensible);
    }
    Type checkBase(Type t,
                   JCTree tree,
                   Env<AttrContext> env,
                   boolean classExpected,
                   boolean interfaceExpected,
                   boolean checkExtensible) {
        if (t.tag == TYPEVAR && !classExpected && !interfaceExpected) {
            // check that type variable is already visible
            if (t.getUpperBound() == null) {
                log.error(tree.pos(), "illegal.forward.ref");
                return types.createErrorType(t);
            }
        } else {
            t = chk.checkClassType(tree.pos(), t, checkExtensible|!allowGenerics);
        }
        if (interfaceExpected && (t.tsym.flags() & INTERFACE) == 0) {
            log.error(tree.pos(), "intf.expected.here");
            // return errType is necessary since otherwise there might
            // be undetected cycles which cause attribution to loop
            return types.createErrorType(t);
        } else if (checkExtensible &&
                   classExpected &&
                   (t.tsym.flags() & INTERFACE) != 0) {
            log.error(tree.pos(), "no.intf.expected.here");
            return types.createErrorType(t);
        }
        if (checkExtensible &&
            ((t.tsym.flags() & FINAL) != 0)) {
            log.error(tree.pos(),
                      "cant.inherit.from.final", t.tsym);
        }
        chk.checkNonCyclic(tree.pos(), t);
        return t;
    }

    public void visitClassDef(JCClassDecl tree) {
        // Local classes have not been entered yet, so we need to do it now:
        if ((env.info.scope.owner.kind & (VAR | MTH)) != 0)
            enter.classEnter(tree, env);

        ClassSymbol c = tree.sym;
        
//        Env<AttrContext> classDefEnv = env;
//        boolean inAnonClass = false;
//        while (classDefEnv.outer != null) {
//            // System.out.println("\t"+classDefEnv.info.scope.owner);
//            if ((classDefEnv.info.scope.owner.kind & TYP) != 0) {
//                // System.out.println("\t\tclass: "+classDefEnv.info.scope.owner);
//                if (classDefEnv.info.scope.owner.name.isEmpty()) {
//                    // System.out.println("\t\t    ==> anonymous class");
//                    inAnonClass = true;
//                }
//            }
//            classDefEnv = classDefEnv.outer;
//        }
        
        if (c == null) {
            // exit in case something drastic went wrong during enter.
            result = null;
        } else {
            // make sure class has been completed:
            c.complete();

            // If this class appears as an anonymous class
            // in a superclass constructor call where
            // no explicit outer instance is given,
            // disable implicit outer instance from being passed.
            // (This would be an illegal access to "this before super").
            if (env.info.isSelfCall &&
                env.tree.getTag() == JCTree.NEWCLASS &&
                ((JCNewClass) env.tree).encl == null)
            {
                c.flags_field |= NOOUTERTHIS;
            }
            attribClass(tree.pos(), c);
            result = tree.type = c.type;
        }
    }

    public void visitMethodDef(JCMethodDecl tree) {
        MethodSymbol m = tree.sym;

        Lint lint = env.info.lint.augment(m.attributes_field, m.flags());
        Lint prevLint = chk.setLint(lint);
        try {
            chk.checkDeprecatedAnnotation(tree.pos(), m);

            attribBounds(tree.typarams);

            // If we override any other methods, check that we do so properly.
            // JLS ???
            chk.checkOverride(tree, m);

            // Create a new environment with local scope
            // for attributing the method.
            Env<AttrContext> localEnv = memberEnter.methodEnv(tree, env);

            localEnv.info.lint = lint;

            // Enter all type parameters into the local method scope.
            for (List<JCTypeParameter> l = tree.typarams; l.nonEmpty(); l = l.tail)
                localEnv.info.scope.enterIfAbsent(l.head.type.tsym);

            ClassSymbol owner = env.enclClass.sym;
            if ((owner.flags() & ANNOTATION) != 0 &&
                tree.params.nonEmpty())
                log.error(tree.params.head.pos(),
                          "intf.annotation.members.cant.have.params");

            // Attribute all value parameters.
            for (List<JCVariableDecl> l = tree.params; l.nonEmpty(); l = l.tail) {
                attribStat(l.head, localEnv);
            }

            // Check that type parameters are well-formed.
            chk.validate(tree.typarams, localEnv);
            if ((owner.flags() & ANNOTATION) != 0 &&
                tree.typarams.nonEmpty())
                log.error(tree.typarams.head.pos(),
                          "intf.annotation.members.cant.have.type.params");

            // Check that result type is well-formed.
            chk.validate(tree.restype, localEnv);
            if ((owner.flags() & ANNOTATION) != 0)
                chk.validateAnnotationType(tree.restype);

            if ((owner.flags() & ANNOTATION) != 0)
                chk.validateAnnotationMethod(tree.pos(), m);

            // Check that all exceptions mentioned in the throws clause extend
            // java.lang.Throwable.
            if ((owner.flags() & ANNOTATION) != 0 && tree.thrown.nonEmpty())
                log.error(tree.thrown.head.pos(),
                          "throws.not.allowed.in.intf.annotation");
            for (List<JCExpression> l = tree.thrown; l.nonEmpty(); l = l.tail)
                chk.checkType(l.head.pos(), l.head.type, syms.throwableType);

            if (tree.body == null) {
                // Empty bodies are only allowed for
                // abstract, native, or interface methods, or for methods
                // in a retrofit signature class.
                if ((owner.flags() & INTERFACE) == 0 &&
                    (tree.mods.flags & (ABSTRACT | NATIVE)) == 0 &&
                    !relax)
                    log.error(tree.pos(), "missing.meth.body.or.decl.abstract");
                if (tree.defaultValue != null) {
                    if ((owner.flags() & ANNOTATION) == 0)
                        log.error(tree.pos(),
                                  "default.allowed.in.intf.annotation.member");
                }
            } else if ((owner.flags() & INTERFACE) != 0) {
                log.error(tree.body.pos(), "intf.meth.cant.have.body");
            } else if ((tree.mods.flags & ABSTRACT) != 0) {
                log.error(tree.pos(), "abstract.meth.cant.have.body");
            } else if ((tree.mods.flags & NATIVE) != 0) {
                log.error(tree.pos(), "native.meth.cant.have.body");
            } else {
                // Add an implicit super() call unless an explicit call to
                // super(...) or this(...) is given
                // or we are compiling class java.lang.Object.
                if (tree.name == names.init && owner.type != syms.objectType) {
                    JCBlock body = tree.body;
                    if (body.stats.isEmpty() ||
                        !TreeInfo.isSelfCall(body.stats.head)) {
                        body.stats = body.stats.
                            prepend(memberEnter.SuperCall(make.at(body.pos),
                                                          List.<Type>nil(),
                                                          List.<JCVariableDecl>nil(),
                                                          false));
                    } else if ((env.enclClass.sym.flags() & ENUM) != 0 &&
                               (tree.mods.flags & GENERATEDCONSTR) == 0 &&
                               TreeInfo.isSuperCall(body.stats.head)) {
                        // enum constructors are not allowed to call super
                        // directly, so make sure there aren't any super calls
                        // in enum constructors, except in the compiler
                        // generated one.
                        log.error(tree.body.stats.head.pos(),
                                  "call.to.super.not.allowed.in.enum.ctor",
                                  env.enclClass.sym);
                    }
                }

                // Attribute method body.
                attribStat(tree.body, localEnv);
            }
            localEnv.info.scope.leave();
            result = tree.type = m.type;
            chk.validateAnnotations(tree.mods.annotations, m);

            if (tree.name == names.init) {
                boolean isCtorSeparable = ((tree.sym.flags() & SEPARABLE)!=0);
                env.enclClass.containsSeparableCtors |= isCtorSeparable; 
            }
//            if (((tree.sym.flags() & SEPARABLE)!=0) && (tree.thrown.size()>0)) {
//                // separable method cannot have a throws clause
//                log.error(tree.thrown.head.pos(), "no.throws.clause.for.separable");
//            }
        }
        finally {
            chk.setLint(prevLint);
        }
    }

    public void visitVarDef(JCVariableDecl tree) {
        // Local variables have not been entered yet, so we need to do it now:
        if (env.info.scope.owner.kind == MTH) {
            if (tree.sym != null) {
                // parameters have already been entered
                env.info.scope.enter(tree.sym);
            } else {
                memberEnter.memberEnter(tree, env);
                annotate.flush();
            }
        }

        VarSymbol v = tree.sym;
        Lint lint = env.info.lint.augment(v.attributes_field, v.flags());
        Lint prevLint = chk.setLint(lint);

        // Check that the variable's declared type is well-formed.
        chk.validate(tree.vartype, env);

        try {
            chk.checkDeprecatedAnnotation(tree.pos(), v);

            if (tree.init != null) {
                if ((v.flags_field & FINAL) != 0 && tree.init.getTag() != JCTree.NEWCLASS) {
                    // In this case, `v' is final.  Ensure that it's initializer is
                    // evaluated.
                    v.getConstValue(); // ensure initializer is evaluated
                } else {
                    // Attribute initializer in a new environment
                    // with the declared variable as owner.
                    // Check that initializer conforms to variable's declared type.
                    Env<AttrContext> initEnv = memberEnter.initEnv(tree, env);
                    initEnv.info.lint = lint;
                    // In order to catch self-references, we set the variable's
                    // declaration position to maximal possible value, effectively
                    // marking the variable as undefined.
                    initEnv.info.enclVar = v;
                    attribExpr(tree.init, initEnv, v.type);
                }
            }
            result = tree.type = v.type;
            chk.validateAnnotations(tree.mods.annotations, v);
        }
        finally {
            chk.setLint(prevLint);
        }
    }

    public void visitSkip(JCSkip tree) {
        result = null;
    }

    public void visitBlock(JCBlock tree) {
        if (env.info.scope.owner.kind == TYP) {
            // Block is a static or instance initializer;
            // let the owner of the environment be a freshly
            // created BLOCK-method.
            Env<AttrContext> localEnv =
                env.dup(tree, env.info.dup(env.info.scope.dupUnshared()));
            localEnv.info.scope.owner =
                new MethodSymbol(tree.flags | BLOCK, names.empty, null,
                                 env.info.scope.owner);
            if ((tree.flags & STATIC) != 0) localEnv.info.staticLevel++;
            attribStats(tree.stats, localEnv);
        } else {
            // Create a new local environment with a local scope.
            Env<AttrContext> localEnv =
                env.dup(tree, env.info.dup(env.info.scope.dup()));
            attribStats(tree.stats, localEnv);
            localEnv.info.scope.leave();
        }
        result = null;
    }

    // mgr: staging additions

    public void visitBracketExpr(JCBracketExpr tree) {
        if (options.get("-verbose") != null) { System.out.println("Working on bracket: "+tree); }
        tree.setSafe(true); // assumed safe until proven guilty
        Env<AttrContext> localEnv = enter.bracketEnv(tree, env);

        attribExpr(tree.body, localEnv);
        tree.classVarsInScope = getClassVarsInScope();
        
        if (options.get("-verbose") != null) { System.out.println("Bracket: "+tree+" is ====="+(tree.isSafe()?"":"UN")+"SAFE====="); }
        // if the type of the body is a primitive type, we want the boxed
        // type instead
        Type t = tree.body.type;
        if (t.isPrimitive()) {
            t = types.boxedClass(t).type;
        }
        if (t.equals(syms.voidType)) {
            log.error(tree.body.pos(), "bracket.expression.cannot.have.type.void");
        }
        // type of brackets is Code<T>, where T is the type of the body
        // or SafeCode<T> if bracket is safe
        Type bracketType = tree.isSafe()?syms.safeCodeType:syms.codeType;
        Type owntype = new ClassType (bracketType.getEnclosingType(),
                                      List.of(t), bracketType.tsym);
        result = check (tree, owntype, pkind, pkind, pt);
    }

    public void visitBracketStat(JCBracketStat tree) {
        tree.setSafe(true); // assumed safe until proven guilty
        // Create a new local environment with a local scope.
        Env<AttrContext> localEnv = enter.bracketEnv(tree, env);
        attribStats(tree.body, localEnv);
        tree.classVarsInScope = getClassVarsInScope();
        if (options.get("-verbose") != null) { System.out.println("Bracket: "+tree+" is ====="+(tree.isSafe()?"":"UN")+"SAFE====="); }
        // localEnv.info.scope.leave();
        // type of brackets is Code<Void> or SafeCode<Void< if bracket is safe
        Type bracketType = tree.isSafe()?syms.safeCodeType:syms.codeType;
        Type owntype = new ClassType (bracketType.getEnclosingType(),
                                      List.of(syms.javaLangVoidType), bracketType.tsym);
        result = check (tree, owntype, pkind, pkind, pt);
    }
    
    /** This method returns a list of all VarSymbols of type Class<X> in scope where X is a type variable. */
    List<VarSymbol> getClassVarsInScope() {
        // System.out.println("env = "+env);
        List<VarSymbol> ret = List.nil();
        Scope current = env.info.scope;
        // System.out.println("Scope: "+current);
        while(current != null) { // also go into the outer (enclosing) scopes
            // for(Symbol s: current.elems) {
            for (Scope.Entry e = current.elems; e != null; e = e.sibling) {
              ret = processPotentialClassVarInScope(e.sym, ret);
            }
            current = current.next;
        }
        // System.out.println("Asking for vars in "+env.enclClass.sym);
        ret = _getClassVarsInScope(env, ret);
        return ret.reverse();
    }
    
    List<VarSymbol> processPotentialClassVarInScope(Symbol s, List<VarSymbol> ret) {
        // System.out.println("in scope: "+s+" : "+s.getClass().getSimpleName());
        if (s instanceof VarSymbol) {
            VarSymbol vs = (VarSymbol)s;
            if ((vs.type.tsym == syms.classType.tsym) &&
                ((vs.flags() & FINAL) != 0) &&
                (vs.type.getTypeArguments().nonEmpty()) &&
                (vs.type.getTypeArguments().head instanceof TypeVar)) {
                // vs is of type Class<X> where X is a type variable
                ret = ret.prepend(vs);
            }
        }
        return ret;
    }
    
    List<VarSymbol> _getClassVarsInScope(Env<AttrContext> env, List<VarSymbol> ret) {
        Env<AttrContext> env1 = env;
        boolean staticOnly = false;
        while (env1.outer != null) {
            if (rs.isStatic(env1)) staticOnly = true;            
            Scope.Entry e = env1.info.scope.elems;
            while (e != null) {
                Symbol sym = e.sym;
                if (!(staticOnly &&
                      sym.kind == VAR &&
                      sym.owner.kind == TYP &&
                      (sym.flags() & STATIC) == 0)) {
                    ret = processPotentialClassVarInScope(sym, ret);
                }
                ret = _getClassFieldsInScope(env1, env1.enclClass.sym.type, env1.enclClass.sym, ret);
                
                e = e.sibling;
            }
            ret = _getClassFieldsInScope(env1, env1.enclClass.sym.type, env1.enclClass.sym, ret);
            
            if ((env1.enclClass.sym.flags() & STATIC) != 0) staticOnly = true;
            env1 = env1.outer;
        }

        ret = _getClassFieldsInScope(env, syms.predefClass.type, syms.predefClass, ret);
        
        return ret;
    }
    
    List<VarSymbol> _getClassFieldsInScope(Env<AttrContext> env,
                                           Type site,
                                           TypeSymbol c,
                                           List<VarSymbol> ret) {
        while (c.type.tag == TYPEVAR)
            c = c.type.getUpperBound().tsym;
        Scope.Entry e = c.members().elems;
        while (e != null) {
            if (e.sym.kind == VAR && (e.sym.flags_field & SYNTHETIC) == 0) {
                if (rs.isAccessible(env, site, e.sym)) {
                    ret = processPotentialClassVarInScope(e.sym, ret);
                }
            }
            e = e.sibling;
        }
        Type st = types.supertype(c.type);
        if (st != null && (st.tag == CLASS || st.tag == TYPEVAR)) {
            ret = _getClassFieldsInScope(env, site, st.tsym, ret);
        }
        for (Type it: types.interfaces(c.type)) {
            ret = _getClassFieldsInScope(env, site, it.tsym, ret);
        }
        return ret;
    }
    
    public void visitEscapeExpr(JCEscapeExpr tree) {
      // check if there are enclosing brackets
      if (env.enclBrackets.isEmpty ()) {
        log.error (tree.body.pos(), "escape.outside.brackets");
      }
      
      Env<AttrContext> localEnv = enter.escapeEnv(tree, env);
      Type t = attribExpr(tree.body, localEnv);
      // System.out.println("visitEscapeExpr: "+tree+", escapeCount: "+env.escapeCount+", inEscape? "+inEscape+", inSeparableMethod? "+inSeparableMethod);
      // System.out.println("\t"+t);      
      // System.out.println("(t.tsym == syms.safeCodeType.tsym) = "+(t.tsym == syms.safeCodeType.tsym));
      // System.out.println("\t(t.tsym == syms.codeType.tsym) = "+(t.tsym == syms.codeType.tsym));
      // System.out.println("\tenv.isSeparable() = "+env.isSeparable());
      // System.out.println("((t.tsym == syms.codeType.tsym) && !env.isSeparable()) = "+((t.tsym == syms.codeType.tsym) && !env.isSeparable()));
      // System.out.println("isClassCloserThanBrackets() = "+isClassCloserThanBrackets());
      
      // if (t instanceof TypeVar) {
      //   TypeVar tv = (TypeVar)t;
      //   System.out.println("typevar: "+tv+", bound: "+tv.bound+" : "+tv.bound.getClass());
      // }
      if ((t instanceof ClassType) &&
          ((t.tsym == syms.safeCodeType.tsym) ||
           ((t.tsym == syms.codeType.tsym) && (!env.isSeparable() || !isClassCloserThanBrackets())))) {
        if (t.isRaw()) {
          log.error (tree.body.pos(), "code.may.not.be.raw");
        }
        else {
          result = check (tree, t.getTypeArguments().head, pkind, pkind, pt);
          if (t.tsym == syms.codeType.tsym) setClosestBracketUnsafe(tree);
        }
      }
      else if ((t instanceof TypeVar) &&
               ((((TypeVar)t).bound.tsym == syms.safeCodeType.tsym) ||
                ((((TypeVar)t).bound.tsym == syms.codeType.tsym) && (!env.isSeparable() || !isClassCloserThanBrackets())))) {
        TypeVar tv = (TypeVar)t;
        if (tv.bound.isRaw()) {
          log.error (tree.body.pos(), "code.may.not.be.raw");
        }
        else {
          result = check (tree, t.getTypeArguments().head, pkind, pkind, pt);
          if (tv.bound.tsym == syms.codeType.tsym) setClosestBracketUnsafe(tree);
        }
      } else {
        // System.out.println("JCEscapeExpr: "+tree);
        // System.out.println("env.isSeparable: "+env.isSeparable());
        // System.out.println("isClassCloserThanBrackets: "+isClassCloserThanBrackets());
        if (env.isSeparable() && isClassCloserThanBrackets()) {
          log.error (tree.body.pos(), "escape.in.separable.needs.safecode.type", t);
        }
        else {
          log.error (tree.body.pos(), "escape.needs.code.type", t);
        }
        result = types.createErrorType(t);
      }
    }
    
    public void visitEscapeStat(JCEscapeStat tree) {
        // check if there are enclosing brackets
        if (env.enclBrackets.isEmpty ()) {
            log.error (tree.body.pos(), "escape.outside.brackets");
        }

        Env<AttrContext> localEnv = enter.escapeEnv(tree, env);
        Type t = attribExpr(tree.body, localEnv);
        
        if (t.isRaw()) {
            log.error (tree.body.pos(), "code.may.not.be.raw");
            result = types.createErrorType(t);
            if (t.tsym == syms.codeType.tsym) setClosestBracketUnsafe(tree);
        }
        else {
            // check for Code<Void> type
            if ((t instanceof ClassType) &&
                ((t.tsym == syms.safeCodeType.tsym) ||
                 ((t.tsym == syms.codeType.tsym) && (!env.isSeparable() || !isClassCloserThanBrackets()))) &&
                (t.getTypeArguments().nonEmpty()) && (t.getTypeArguments().head.tsym == syms.javaLangVoidType.tsym)) {
                result = tree.type = t.getParameterTypes().head;
            }
            else if ((t instanceof ClassType) &&
                     ((t.tsym == syms.safeCodeType.tsym) ||
                      ((t.tsym == syms.codeType.tsym) && (!env.isSeparable() || !isClassCloserThanBrackets()))) &&
                     (t.getTypeArguments().nonEmpty()) && (t.getTypeArguments().head instanceof Type.CapturedType)) {
                // warning for Code<?> or SafeCode<?> escape in a statement
//                if (env.isSeparable()) {
//                    log.warning (tree.body.pos(), "escape.stat.in.separable.safecode.wildcard.type");
//                }
//                else {
//                    log.warning (tree.body.pos(), "escape.stat.code.wildcard.type");
//                }
//                result = tree.type = t.getParameterTypes().head;
                // error for Code<?> or SafeCode<?> escape in a statement
                // if (env.isSeparable() && !isClassCloserThanBrackets()) {
                if (t.tsym == syms.safeCodeType.tsym) {
                    log.error (tree.body.pos(), "escape.stat.in.separable.safecode.wildcard.type");
                }
                else {
                    log.error (tree.body.pos(), "escape.stat.code.wildcard.type");
                }
                result = types.createErrorType(t);
            }
            // TODO: TypeVar (T extends Code<Void>)
            else {
                // warning for non-void escape in a statement
                if (env.isSeparable() && isClassCloserThanBrackets()) {
                    if (t.tsym != syms.safeCodeType.tsym) {
                        log.error (tree.body.pos(), "escape.stat.in.separable.safecode.wildcard.type");
                    }
                    else {
                        log.warning (tree.body.pos(), "escape.stat.in.separable.not.safecode.void.type", t);
                    }
                }
                else {
                    log.warning (tree.body.pos(), "escape.stat.not.code.void.type", t);
                }
                result = tree.type = t.getParameterTypes().head;
            }
        }
    }

    public void visitDoLoop(JCDoWhileLoop tree) {
        attribStat(tree.body, env.dup(tree));
        attribExpr(tree.cond, env, syms.booleanType);
        result = null;
    }

    public void visitWhileLoop(JCWhileLoop tree) {
        attribExpr(tree.cond, env, syms.booleanType);
        attribStat(tree.body, env.dup(tree));
        result = null;
    }

    public void visitForLoop(JCForLoop tree) {
        Env<AttrContext> loopEnv =
            env.dup(env.tree, env.info.dup(env.info.scope.dup()));
        attribStats(tree.init, loopEnv);
        if (tree.cond != null) attribExpr(tree.cond, loopEnv, syms.booleanType);
        loopEnv.tree = tree; // before, we were not in loop!
        attribStats(tree.step, loopEnv);
        attribStat(tree.body, loopEnv);
        loopEnv.info.scope.leave();
        result = null;
    }

    public void visitForeachLoop(JCEnhancedForLoop tree) {
        Env<AttrContext> loopEnv =
            env.dup(env.tree, env.info.dup(env.info.scope.dup()));
        attribStat(tree.var, loopEnv);
        Type exprType = types.upperBound(attribExpr(tree.expr, loopEnv));
        chk.checkNonVoid(tree.pos(), exprType);
        Type elemtype = types.elemtype(exprType); // perhaps expr is an array?
        if (elemtype == null) {
            // or perhaps expr implements Iterable<T>?
            Type base = types.asSuper(exprType, syms.iterableType.tsym);
            if (base == null) {
                log.error(tree.expr.pos(), "foreach.not.applicable.to.type");
                elemtype = types.createErrorType(exprType);
            } else {
                List<Type> iterableParams = base.allparams();
                elemtype = iterableParams.isEmpty()
                    ? syms.objectType
                    : types.upperBound(iterableParams.head);
            }
        }
        chk.checkType(tree.expr.pos(), elemtype, tree.var.sym.type);
        loopEnv.tree = tree; // before, we were not in loop!
        attribStat(tree.body, loopEnv);
        loopEnv.info.scope.leave();
        result = null;
    }

    public void visitLabelled(JCLabeledStatement tree) {
        // Check that label is not used in an enclosing statement
        Env<AttrContext> env1 = env;
        while (env1 != null && env1.tree.getTag() != JCTree.CLASSDEF) {
            if (env1.tree.getTag() == JCTree.LABELLED &&
                ((JCLabeledStatement) env1.tree).label == tree.label) {
                log.error(tree.pos(), "label.already.in.use",
                          tree.label);
                break;
            }
            env1 = env1.next;
        }

        attribStat(tree.body, env.dup(tree));
        result = null;
    }

    public void visitSwitch(JCSwitch tree) {
        Type seltype = attribExpr(tree.selector, env);

        Env<AttrContext> switchEnv =
            env.dup(tree, env.info.dup(env.info.scope.dup()));

        boolean enumSwitch =
            allowEnums &&
            (seltype.tsym.flags() & Flags.ENUM) != 0;
        if (!enumSwitch)
            seltype = chk.checkType(tree.selector.pos(), seltype, syms.intType);

        // Attribute all cases and
        // check that there are no duplicate case labels or default clauses.
        Set<Object> labels = new HashSet<Object>(); // The set of case labels.
        boolean hasDefault = false;      // Is there a default label?
        for (List<JCCase> l = tree.cases; l.nonEmpty(); l = l.tail) {
            JCCase c = l.head;
            Env<AttrContext> caseEnv =
                switchEnv.dup(c, env.info.dup(switchEnv.info.scope.dup()));
            if (c.pat != null) {
                if (enumSwitch) {
                    Symbol sym = enumConstant(c.pat, seltype);
                    if (sym == null) {
                        log.error(c.pat.pos(), "enum.const.req");
                    } else if (!labels.add(sym)) {
                        log.error(c.pos(), "duplicate.case.label");
                    }
                } else {
                    Type pattype = attribExpr(c.pat, switchEnv, seltype);
                    if (pattype.tag != ERROR) {
                        if (pattype.constValue() == null) {
                            log.error(c.pat.pos(), "const.expr.req");
                        } else if (labels.contains(pattype.constValue())) {
                            log.error(c.pos(), "duplicate.case.label");
                        } else {
                            labels.add(pattype.constValue());
                        }
                    }
                }
            } else if (hasDefault) {
                log.error(c.pos(), "duplicate.default.label");
            } else {
                hasDefault = true;
            }
            attribStats(c.stats, caseEnv);
            caseEnv.info.scope.leave();
            addVars(c.stats, switchEnv.info.scope);
        }

        switchEnv.info.scope.leave();
        result = null;
    }
    // where
        /** Add any variables defined in stats to the switch scope. */
        private static void addVars(List<JCStatement> stats, Scope switchScope) {
            for (;stats.nonEmpty(); stats = stats.tail) {
                JCTree stat = stats.head;
                if (stat.getTag() == JCTree.VARDEF)
                    switchScope.enter(((JCVariableDecl) stat).sym);
            }
        }
    // where
    /** Return the selected enumeration constant symbol, or null. */
    private Symbol enumConstant(JCTree tree, Type enumType) {
        if (tree.getTag() != JCTree.IDENT) {
            log.error(tree.pos(), "enum.label.must.be.unqualified.enum");
            return syms.errSymbol;
        }
        JCIdent ident = (JCIdent)tree;
        Name name = ident.name;
        for (Scope.Entry e = enumType.tsym.members().lookup(name);
             e.scope != null; e = e.next()) {
            if (e.sym.kind == VAR) {
                Symbol s = ident.sym = e.sym;
                ((VarSymbol)s).getConstValue(); // ensure initializer is evaluated
                ident.type = s.type;
                return ((s.flags_field & Flags.ENUM) == 0)
                    ? null : s;
            }
        }
        return null;
    }

    public void visitSynchronized(JCSynchronized tree) {
        chk.checkRefType(tree.pos(), attribExpr(tree.lock, env));
        attribStat(tree.body, env);
        result = null;
    }

    public void visitTry(JCTry tree) {
        // Attribute body
        attribStat(tree.body, env.dup(tree, env.info.dup()));

        // Attribute catch clauses
        for (List<JCCatch> l = tree.catchers; l.nonEmpty(); l = l.tail) {
            JCCatch c = l.head;
            Env<AttrContext> catchEnv =
                env.dup(c, env.info.dup(env.info.scope.dup()));
            Type ctype = attribStat(c.param, catchEnv);
            if (c.param.type.tsym.kind == Kinds.VAR) {
                c.param.sym.setData(ElementKind.EXCEPTION_PARAMETER);
            }
            chk.checkType(c.param.vartype.pos(),
                          chk.checkClassType(c.param.vartype.pos(), ctype),
                          syms.throwableType);
            attribStat(c.body, catchEnv);
            catchEnv.info.scope.leave();
        }

        // Attribute finalizer
        if (tree.finalizer != null) attribStat(tree.finalizer, env);
        result = null;
    }

    public void visitConditional(JCConditional tree) {
        attribExpr(tree.cond, env, syms.booleanType);
        attribExpr(tree.truepart, env);
        attribExpr(tree.falsepart, env);
        result = check(tree,
                       capture(condType(tree.pos(), tree.cond.type,
                                        tree.truepart.type, tree.falsepart.type)),
                       VAL, pkind, pt);
    }
    //where
        /** Compute the type of a conditional expression, after
         *  checking that it exists. See Spec 15.25.
         *
         *  @param pos      The source position to be used for
         *                  error diagnostics.
         *  @param condtype The type of the expression's condition.
         *  @param thentype The type of the expression's then-part.
         *  @param elsetype The type of the expression's else-part.
         */
        private Type condType(DiagnosticPosition pos,
                              Type condtype,
                              Type thentype,
                              Type elsetype) {
            Type ctype = condType1(pos, condtype, thentype, elsetype);

            // If condition and both arms are numeric constants,
            // evaluate at compile-time.
            return ((condtype.constValue() != null) &&
                    (thentype.constValue() != null) &&
                    (elsetype.constValue() != null))
                ? cfolder.coerce(condtype.isTrue()?thentype:elsetype, ctype)
                : ctype;
        }
        /** Compute the type of a conditional expression, after
         *  checking that it exists.  Does not take into
         *  account the special case where condition and both arms
         *  are constants.
         *
         *  @param pos      The source position to be used for error
         *                  diagnostics.
         *  @param condtype The type of the expression's condition.
         *  @param thentype The type of the expression's then-part.
         *  @param elsetype The type of the expression's else-part.
         */
        private Type condType1(DiagnosticPosition pos, Type condtype,
                               Type thentype, Type elsetype) {
            // If same type, that is the result
            if (types.isSameType(thentype, elsetype))
                return thentype.baseType();

            Type thenUnboxed = (!allowBoxing || thentype.isPrimitive())
                ? thentype : types.unboxedType(thentype);
            Type elseUnboxed = (!allowBoxing || elsetype.isPrimitive())
                ? elsetype : types.unboxedType(elsetype);

            // Otherwise, if both arms can be converted to a numeric
            // type, return the least numeric type that fits both arms
            // (i.e. return larger of the two, or return int if one
            // arm is short, the other is char).
            if (thenUnboxed.isPrimitive() && elseUnboxed.isPrimitive()) {
                // If one arm has an integer subrange type (i.e., byte,
                // short, or char), and the other is an integer constant
                // that fits into the subrange, return the subrange type.
                if (thenUnboxed.tag < INT && elseUnboxed.tag == INT &&
                    types.isAssignable(elseUnboxed, thenUnboxed))
                    return thenUnboxed.baseType();
                if (elseUnboxed.tag < INT && thenUnboxed.tag == INT &&
                    types.isAssignable(thenUnboxed, elseUnboxed))
                    return elseUnboxed.baseType();

                for (int i = BYTE; i < VOID; i++) {
                    Type candidate = syms.typeOfTag[i];
                    if (types.isSubtype(thenUnboxed, candidate) &&
                        types.isSubtype(elseUnboxed, candidate))
                        return candidate;
                }
            }

            // Those were all the cases that could result in a primitive
            if (allowBoxing) {
                if (thentype.isPrimitive())
                    thentype = types.boxedClass(thentype).type;
                if (elsetype.isPrimitive())
                    elsetype = types.boxedClass(elsetype).type;
            }

            if (types.isSubtype(thentype, elsetype))
                return elsetype.baseType();
            if (types.isSubtype(elsetype, thentype))
                return thentype.baseType();

            if (!allowBoxing || thentype.tag == VOID || elsetype.tag == VOID) {
                log.error(pos, "neither.conditional.subtype",
                          thentype, elsetype);
                return thentype.baseType();
            }

            // both are known to be reference types.  The result is
            // lub(thentype,elsetype). This cannot fail, as it will
            // always be possible to infer "Object" if nothing better.
            return types.lub(thentype.baseType(), elsetype.baseType());
        }

    public void visitIf(JCIf tree) {
        attribExpr(tree.cond, env, syms.booleanType);
        attribStat(tree.thenpart, env);
        if (tree.elsepart != null)
            attribStat(tree.elsepart, env);
        chk.checkEmptyIf(tree);
        result = null;
    }

    public void visitExec(JCExpressionStatement tree) {
        attribExpr(tree.expr, env);
        result = null;
    }

    public void visitBreak(JCBreak tree) {
        tree.target = findJumpTarget(tree.pos(), tree.getTag(), tree.label, env);
        result = null;
    }

    public void visitContinue(JCContinue tree) {
        tree.target = findJumpTarget(tree.pos(), tree.getTag(), tree.label, env);
        result = null;
    }
    //where
        /** Return the target of a break or continue statement, if it exists,
         *  report an error if not.
         *  Note: The target of a labelled break or continue is the
         *  (non-labelled) statement tree referred to by the label,
         *  not the tree representing the labelled statement itself.
         *
         *  @param pos     The position to be used for error diagnostics
         *  @param tag     The tag of the jump statement. This is either
         *                 Tree.BREAK or Tree.CONTINUE.
         *  @param label   The label of the jump statement, or null if no
         *                 label is given.
         *  @param env     The environment current at the jump statement.
         */
        private JCTree findJumpTarget(DiagnosticPosition pos,
                                    int tag,
                                    Name label,
                                    Env<AttrContext> env) {
            // Search environments outwards from the point of jump.
            Env<AttrContext> env1 = env;
            LOOP:
            while (env1 != null) {
                switch (env1.tree.getTag()) {
                case JCTree.LABELLED:
                    JCLabeledStatement labelled = (JCLabeledStatement)env1.tree;
                    if (label == labelled.label) {
                        // If jump is a continue, check that target is a loop.
                        if (tag == JCTree.CONTINUE) {
                            if (labelled.body.getTag() != JCTree.DOLOOP &&
                                labelled.body.getTag() != JCTree.WHILELOOP &&
                                labelled.body.getTag() != JCTree.FORLOOP &&
                                labelled.body.getTag() != JCTree.FOREACHLOOP)
                                log.error(pos, "not.loop.label", label);
                            // Found labelled statement target, now go inwards
                            // to next non-labelled tree.
                            return TreeInfo.referencedStatement(labelled);
                        } else {
                            return labelled;
                        }
                    }
                    break;
                case JCTree.DOLOOP:
                case JCTree.WHILELOOP:
                case JCTree.FORLOOP:
                case JCTree.FOREACHLOOP:
                    if (label == null) return env1.tree;
                    break;
                case JCTree.SWITCH:
                    if (label == null && tag == JCTree.BREAK) return env1.tree;
                    break;
                case JCTree.METHODDEF:
                case JCTree.CLASSDEF:
                    break LOOP;
                default:
                }
                env1 = env1.next;
            }
            if (label != null)
                log.error(pos, "undef.label", label);
            else if (tag == JCTree.CONTINUE)
                log.error(pos, "cont.outside.loop");
            else
                log.error(pos, "break.outside.switch.loop");
            return null;
        }

    public void visitReturn(JCReturn tree) {
        // Check that there is an enclosing method which is
        // nested within than the enclosing class.
        if (env.enclMethod == null ||
            env.enclMethod.sym.owner != env.enclClass.sym) {
            log.error(tree.pos(), "ret.outside.meth");

        } else {
            // Attribute return expression, if it exists, and check that
            // it conforms to result type of enclosing method.
            Symbol m = env.enclMethod.sym;
            if (m.type.getReturnType().tag == VOID) {
                if (tree.expr != null)
                    log.error(tree.expr.pos(),
                              "cant.ret.val.from.meth.decl.void");
            } else if (tree.expr == null) {
                log.error(tree.pos(), "missing.ret.val");
            } else {
                attribExpr(tree.expr, env, m.type.getReturnType());
            }
        }
        
        if (!isClassCloserThanBrackets()) {
            log.error(tree.pos(), "no.return.in.brackets");            
        }
        result = null;
    }
    
    protected boolean checkThrowSeparable(JCExpression excExpr) {
        // System.out.println("checkThrowSeparable: "+excExpr.getClass()+" = "+excExpr+" : "+excExpr.type);
        if (!(excExpr instanceof JCNewClass)) {
            // not a `new C(e1, e2, ..., en)' construct
            return false;
        }
        Type t = excExpr.type;
        if ((t!=null) && (t.tsym!=null) && (t.tsym.name.isEmpty())) {
            // System.out.println("Anonymous: "+t);
            return false;
        }

        JCNewClass ncExpr = (JCNewClass)excExpr;
        for(JCExpression arg: ncExpr.args) {
            // System.out.println("\targ "+arg+" code-free? "+arg.type);
            // System.out.println("\t\t"+checkTypeCodeFree(arg.type));
            if (!checkTypeCodeFree(arg.type)) return false;
        }
        return true;
    }

    public void visitThrow(JCThrow tree) {
        attribExpr(tree.expr, env, syms.throwableType);
        result = null;
        // mgr: escape safety
        // in an escape, you can only throw exceptions
        // - whose class is RuntimeException or a subclass
        // - that are thrown using a `throw new C(e1, e2, ..., en)' construct (in tree.expr) where ei are code-free
        // - it is not an anonymous inner class
        if (tree.expr.type.tag==TypeTags.BOT) { return; } // throw null is allowed
        if (!checkThrowSeparable(tree.expr)) {
            // TODO: check if this exception is caught before it leaves the term (escape or separable)! if so, this throw is allowed!
            if (env.enclMethod==null) {
                // throw in an initializer
                // this is an an error if we have any separable contructors
                env.enclClass.containsNonSeparableInitializers = true;
            }
            else {
                if (env.isSeparable()) {
                    // error: throw in an separable method or in an escape
                    log.error(tree.pos(), "no.unsafe.throw.in.separable");                
                }
            }
        }
    }

    public void visitAssert(JCAssert tree) {
        attribExpr(tree.cond, env, syms.booleanType);
        if (tree.detail != null) {
            chk.checkNonVoid(tree.detail.pos(), attribExpr(tree.detail, env));
        }
        result = null;
    }

    /** Return true if we are inside brackets, but there is an anonymous inner class even closer:
      * <| ... new Foo() { HERE } ... |>
      * Also returns true if we aren't inside brackets at all.
      */
    protected boolean isClassCloserThanBrackets() {
        if (env.enclBrackets.size()==0) return true; 
        JCExpression head = env.enclBrackets.head;
        
        // find the environment where the bracket was created
        Env<AttrContext> bracketEnclEnv = null;
        Env<AttrContext> e = env;
        while(e!=null) {
            if (e.tree==head) {
                bracketEnclEnv = e;
                break;
            }
            e = e.next;
        }
        // see if we get to env.enclClass before we get to bracketEnclEnv
        boolean found = false;
        e = env;
        while(e!=null) {
            if (e==bracketEnclEnv) {
                break;
            }
            if ((e.info.scope.owner instanceof ClassSymbol) &&
                (((ClassSymbol)e.info.scope.owner)==env.enclClass.sym)) {
                found = true;
                break;
            }
            e = e.next;
        }
        return found; // method closer
    }
    
    protected void setClosestBracketUnsafe(JCTree tree) {
        if (env.enclBrackets.size()>0) {
            if (isClassCloserThanBrackets()) return; // inner class, this is safe

            if (options.get("-verbose") != null) { System.out.println("Unsafe: "+tree); }
            
            JCExpression head = env.enclBrackets.head;
            if (head instanceof JCBracketExpr) {
                ((JCBracketExpr)head).safe = false;
            }
            else if (head instanceof JCBracketStat) {
                ((JCBracketStat)head).safe = false;
            }
            else {
                throw new AssertionError("Expected JCBracketExpr or JCBracketStat in enclBrackets, was "+head.getClass().getSimpleName());
            }
        }
    }
    
    /** Visitor method for method invocations.
     *  NOTE: The method part of an application will have in its type field
     *        the return type of the method, not the method's type itself!
     */
    public void visitApply(JCMethodInvocation tree) {
        // The local environment of a method application is
        // a new environment nested in the current one.
        Env<AttrContext> localEnv = env.dup(tree, env.info.dup());

        // The types of the actual method arguments.
        List<Type> argtypes = null;

        // The types of the actual method type arguments.
        List<Type> typeargtypes = null;

        Name methName = TreeInfo.name(tree.meth);
        Type site = null; // non-null only iff isConstructorCall==true

        boolean isConstructorCall =
            methName == names._this || methName == names._super;
    
        if (isConstructorCall) {
            // We are seeing a ...this(...) or ...super(...) call.
            // Check that this is the first statement in a constructor.
            if (checkFirstConstructorStat(tree, env)) {
                // Record the fact
                // that this is a constructor call (using isSelfCall).
                localEnv.info.isSelfCall = true;

                // Attribute arguments, yielding list of argument types.
                argtypes = attribArgs(tree.args, localEnv);
                typeargtypes = attribTypes(tree.typeargs, localEnv);

                // Variable `site' points to the class in which the called
                // constructor is defined.
                site = env.enclClass.sym.type;
                if (methName == names._super) {
                    if (site == syms.objectType) {
                        log.error(tree.meth.pos(), "no.superclass", site);
                        site = types.createErrorType(syms.objectType);
                    } else {
                        site = types.supertype(site);
                    }
                }

                if (site.tag == CLASS) {
                    if (site.getEnclosingType().tag == CLASS) {
                        // we are calling a nested class

                        if (tree.meth.getTag() == JCTree.SELECT) {
                            JCTree qualifier = ((JCFieldAccess) tree.meth).selected;

                            // We are seeing a prefixed call, of the form
                            //     <expr>.super(...).
                            // Check that the prefix expression conforms
                            // to the outer instance type of the class.
                            chk.checkRefType(qualifier.pos(),
                                             attribExpr(qualifier, localEnv,
                                                        site.getEnclosingType()));
                        } else if (methName == names._super) {
                            // qualifier omitted; check for existence
                            // of an appropriate implicit qualifier.
                            rs.resolveImplicitThis(tree.meth.pos(),
                                                   localEnv, site);
                        }
                    } else if (tree.meth.getTag() == JCTree.SELECT) {
                        log.error(tree.meth.pos(), "illegal.qual.not.icls",
                                  site.tsym);
                    }

                    // if we're calling a java.lang.Enum constructor,
                    // prefix the implicit String and int parameters
                    if (site.tsym == syms.enumSym && allowEnums)
                        argtypes = argtypes.prepend(syms.intType).prepend(syms.stringType);

                    // Resolve the called constructor under the assumption
                    // that we are referring to a superclass instance of the
                    // current instance (JLS ???).
                    boolean selectSuperPrev = localEnv.info.selectSuper;
                    localEnv.info.selectSuper = true;
                    localEnv.info.varArgs = false;
                    Symbol sym = rs.resolveConstructor(
                        tree.meth.pos(), localEnv, site, argtypes, typeargtypes);
                    localEnv.info.selectSuper = selectSuperPrev;

                    // Set method symbol to resolved constructor...
                    TreeInfo.setSymbol(tree.meth, sym);

                    // ...and check that it is legal in the current context.
                    // (this will also set the tree's type)
                    Type mpt = newMethTemplate(argtypes, typeargtypes);
                    checkId(tree.meth, site, sym, localEnv, MTH,
                            mpt, tree.varargsElement != null);
                }
                // Otherwise, `site' is an error type and we do nothing
            }
            result = tree.type = syms.voidType;
        } else {
            // Otherwise, we are seeing a regular method call.
            // Attribute the arguments, yielding list of argument types, ...
            argtypes = attribArgs(tree.args, localEnv);
            typeargtypes = attribTypes(tree.typeargs, localEnv);

            // ... and attribute the method using as a prototype a methodtype
            // whose formal argument types is exactly the list of actual
            // arguments (this will also set the method symbol).
            Type mpt = newMethTemplate(argtypes, typeargtypes);
            localEnv.info.varArgs = false;
            Type mtype = attribExpr(tree.meth, localEnv, mpt);
            if (localEnv.info.varArgs)
                assert mtype.isErroneous() || tree.varargsElement != null;

            // Compute the result type.
            Type restype = mtype.getReturnType();
            assert restype.tag != WILDCARD : mtype;

            // as a special case, array.clone() has a result that is
            // the same as static type of the array being cloned
            if (tree.meth.getTag() == JCTree.SELECT &&
                allowCovariantReturns &&
                methName == names.clone &&
                types.isArray(((JCFieldAccess) tree.meth).selected.type))
                restype = ((JCFieldAccess) tree.meth).selected.type;

            // as a special case, x.getClass() has type Class<? extends |X|>
            if (allowGenerics &&
                methName == names.getClass && tree.args.isEmpty()) {
                Type qualifier = (tree.meth.getTag() == JCTree.SELECT)
                    ? ((JCFieldAccess) tree.meth).selected.type
                    : env.enclClass.sym.type;
                restype = new
                    ClassType(restype.getEnclosingType(),
                              List.<Type>of(new WildcardType(types.erasure(qualifier),
                                                               BoundKind.EXTENDS,
                                                               syms.boundClass)),
                              restype.tsym);
            }

            // Check that value of resulting type is admissible in the
            // current context.  Also, capture the return type
            result = check(tree, capture(restype), VAL, pkind, pt);
        }
        
        // mgr: escape safety check
        // the meth AST is either a JCIdent (for calls within the same class) or a JCFieldAccess
        // System.out.println("visitApply: "+tree+" isCtor? "+isConstructorCall);
        boolean calledMethodSeparable = false;
        if (tree.meth instanceof JCIdent) {
            JCIdent ident = (JCIdent)tree.meth;
            if (isConstructorCall && (ident.sym instanceof MethodSymbol) &&
                (((((MethodSymbol)ident.sym).owner.toString().equals("java.lang.Object")) &&
                  (((MethodSymbol)ident.sym).type.toString().equals("()void")) &&
                  (((MethodSymbol)ident.sym).name.toString().equals("<init>"))) ||
                 ((((MethodSymbol)ident.sym).owner.toString().equals("java.lang.RuntimeException")) && // allow RuntimeException
                  (((MethodSymbol)ident.sym).name.toString().equals("<init>"))) ||
                 ((((MethodSymbol)ident.sym).owner.toString().equals("edu.rice.cs.mint.runtime.ClassCode")) && // allow ClassCode
                  (((MethodSymbol)ident.sym).name.toString().equals("<init>"))) ||
                 ((((MethodSymbol)ident.sym).owner.toString().equals("edu.rice.cs.mint.runtime.FieldCode")) && // allow FieldCode
                  (((MethodSymbol)ident.sym).name.toString().equals("<init>"))))) {
                calledMethodSeparable = true;
            }
            else if (isConstructorCall && (ident.sym instanceof MethodSymbol) &&
                ((ident.sym.flags() & GENERATEDCONSTR)!=0)) {
                // generated ctor
                calledMethodSeparable = checkDefaultCtorsSeparable((MethodSymbol)ident.sym);
            }
            else if (!isConstructorCall && (ident.sym instanceof MethodSymbol) &&
                     ((((MethodSymbol)ident.sym).name.toString().equals("getClass")) && // allow final getClass() method
                      (((MethodSymbol)ident.sym).params().isEmpty()))) {
                calledMethodSeparable = true;
            }
            else {
                calledMethodSeparable = ((ident.sym.flags() & SEPARABLE)!=0);
            }
            // System.out.println("\ttree.meth.sym = "+ident.sym+" separable? "+calledMethodSeparable+" generated? "+((ident.sym.flags() & GENERATEDCONSTR)!=0));
            // System.out.println("\t"+((MethodSymbol)ident.sym).owner+" "+((MethodSymbol)ident.sym).type+" "+((MethodSymbol)ident.sym).name);
        }
        else if (tree.meth instanceof JCFieldAccess) {
            JCFieldAccess fa = (JCFieldAccess)tree.meth;
            if ((fa.selected instanceof JCIdent) && (((JCIdent)fa.selected).name==names._this)) {
//                if ((fa.sym.flags() & STATIC)!=0) {
//                    System.out.println("\t\tstatic");
//                    // this.x access to a static x
                    calledMethodSeparable = ((fa.sym.flags() & SEPARABLE)!=0);
//                }
            }
            else if ((fa.sym instanceof MethodSymbol) &&
                     ((((MethodSymbol)fa.sym).owner.toString().equals("edu.rice.cs.mint.runtime.ClassCode")) ||
                      (((MethodSymbol)fa.sym).owner.toString().equals("edu.rice.cs.mint.runtime.FieldCode")))) {
                calledMethodSeparable = true;
            }
            else if ((fa.sym instanceof MethodSymbol) &&
                     ((((MethodSymbol)fa.sym).name.toString().equals("getClass")) && // allow final getClass() method
                      (((MethodSymbol)fa.sym).params().isEmpty()))) {
                calledMethodSeparable = true;
            }
            else {
                // not a this.x access and not a call to our reflection code
                calledMethodSeparable = ((fa.sym.flags() & SEPARABLE)!=0);
            }
            // System.out.println("\ttree.meth.sym = "+fa.sym+" separable? "+calledMethodSeparable+" generated? "+((fa.sym.flags() & GENERATEDCONSTR)!=0));
        }
        else {
            throw new AssertionError("Expected JCIdent or JCFieldAccess in method invocation, was "+tree.meth.getClass().getSimpleName());
        }
        
        if (!calledMethodSeparable) setClosestBracketUnsafe(tree);
        
        // only emit "cannot call not separable from escape/separable" errors if we are in a method, not in brackets
        // if we are in brackets, this would not be an error, it would just make code be Code<T>, not SafeCode<T>
        // System.out.println("visitApply: JCMethodInvocation "+tree);
        if (isClassCloserThanBrackets()) {
            // System.out.println("isClassCloserThanBrackets = true");
            if (env.enclMethod==null) {
                // method call in an initializer block
                // System.out.println("env.enclMethod==null");
                if (!calledMethodSeparable) {
                    // System.out.println("calledMethodSeparable = false");
                    setClosestBracketUnsafe(tree);
                    if (env.escape!=null) {
                        // error: calling a non-separable method from inside an escape in an initializer block
                        // System.out.println("env.escape!=null");
                        String callee = TreeInfo.name(tree.meth).toString();
                        if (isConstructorCall) {
                            callee = site.toString();
                        }
                        StringBuilder sb = new StringBuilder(callee);
                        sb.append('(');
                        if (argtypes!=null) {
                            boolean first = true;
                            for(Type t: argtypes) {
                                if (first) first = false;
                                else sb.append(',');
                                sb.append(t);
                            }
                        }
                        sb.append(')');
                        callee = sb.toString();
                        log.error(tree.pos(),"cannot.call.not.separable.in.escape", callee);
                    }
                    else {
                        // maybe error, depending on constructors: calling a non-separable method from an initializer block
                        // System.out.println("env.escape==null");
                        env.enclClass.containsNonSeparableInitializers = true;
                    }
                }
            }
            else {
                // System.out.println("env.enclMethod=="+env.enclMethod);
                boolean currentMethodSeparable = ((env.enclMethod.sym.flags() & SEPARABLE)!=0);
                // System.out.println("currentMethodSeparable = "+currentMethodSeparable);
                // System.out.println("getLevel = "+env.getLevel());
                if (env.isSeparable ()) {
                    // Checking currentMethodSeparable to disallow Calls to Non-Separable Methods in Separable Code
                    // while env.isSeparable() only checks at level 0
                    // System.out.println("env.isSeparable = true");
                    if (!calledMethodSeparable &&
                        (site != syms.objectType)) { // allow super call to Object
                        // System.out.println("!calledMethodSeparable && (site != syms.objectType)");
                        String callee = TreeInfo.name(tree.meth).toString();
                        if (isConstructorCall) {
                            callee = site.toString();
                        }
                        StringBuilder sb = new StringBuilder(callee);
                        sb.append('(');
                        if (argtypes!=null) {
                            boolean first = true;
                            for(Type t: argtypes) {
                                if (first) first = false;
                                else sb.append(',');
                                sb.append(t);
                            }
                        }
                        sb.append(')');
                        callee = sb.toString();
                        String caller = env.enclMethod.name.toString();
                        if (env.enclMethod.name==names.init) {
                            caller = env.enclClass.sym.toString();
                        }
                        if (env.escape==null) {
                            // within separable method
                            sb.setLength(0);
                            sb.append(caller);
                            sb.append('(');
                            boolean first = true;
                            for(JCVariableDecl decl: env.enclMethod.params) {
                                if (first) first = false;
                                else sb.append(',');
                                sb.append(decl.vartype);
                            }
                            sb.append(')');
                            caller = sb.toString();
                            log.error(tree.pos(),"not.separable.called.from.separable", callee, caller);
                        }
                        else {
                            // within escape
                            log.error(tree.pos(),"cannot.call.not.separable.in.escape", callee);
                        }
                    }
                }
            }
        }
        // mgr: csp safety
        if (env.enclBrackets.size()>0) {
            boolean nonStaticCall = false;
            String method = null;
            if (tree.meth instanceof JCIdent) {
                JCIdent ident = (JCIdent)tree.meth;
                if ((ident.name!=names._this) && (ident.name!=names._super)) {
                    nonStaticCall = ((ident.sym.flags() & STATIC)==0);
                    method = ident.name.toString();
                }
            }
            else if (tree.meth instanceof JCFieldAccess) {
                JCFieldAccess fa = (JCFieldAccess)tree.meth;
                Type t = env.enclClass.sym.type;
                while(t!=null) {
                    if (fa.sym.owner.type.toString().equals(t.toString())) {
                        nonStaticCall = ((fa.sym.flags() & STATIC)==0);
                        method = fa.sym.name.toString();
                        break;
                    }
                    t = t.getEnclosingType();
                }
            }
            else {
                throw new AssertionError("Expected JCIdent or JCFieldAccess in method invocation, was "+tree.meth.getClass().getSimpleName());
            }

            if (nonStaticCall) {
                log.error(tree.pos(), "may.not.directly.access.enclosing", method);
            }
        }

        // System.out.println("\tenv.enclMethod = "+env.enclMethod+" separable? "+currentMethodSeparable);
        chk.validate(tree.typeargs, localEnv);
    }
    //where
        /** Check that given application node appears as first statement
         *  in a constructor call.
         *  @param tree   The application node
         *  @param env    The environment current at the application.
         */
        boolean checkFirstConstructorStat(JCMethodInvocation tree, Env<AttrContext> env) {
            JCMethodDecl enclMethod = env.enclMethod;
            if (enclMethod != null && enclMethod.name == names.init) {
                JCBlock body = enclMethod.body;
                if (body.stats.head.getTag() == JCTree.EXEC &&
                    ((JCExpressionStatement) body.stats.head).expr == tree)
                    return true;
            }
            log.error(tree.pos(),"call.must.be.first.stmt.in.ctor",
                      TreeInfo.name(tree.meth));
            return false;
        }

        /** Obtain a method type with given argument types.
         */
        Type newMethTemplate(List<Type> argtypes, List<Type> typeargtypes) {
            MethodType mt = new MethodType(argtypes, null, null, syms.methodClass);
            return (typeargtypes == null) ? mt : (Type)new ForAll(typeargtypes, mt);
        }

    public void visitNewClass(JCNewClass tree) {
        Type owntype = types.createErrorType(tree.type);

        // The local environment of a class creation is
        // a new environment nested in the current one.
        Env<AttrContext> localEnv = env.dup(tree, env.info.dup());

        // The anonymous inner class definition of the new expression,
        // if one is defined by it.
        JCClassDecl cdef = tree.def;

        // If enclosing class is given, attribute it, and
        // complete class name to be fully qualified
        JCExpression clazz = tree.clazz; // Class field following new
        JCExpression clazzid =          // Identifier in class field
            (clazz.getTag() == JCTree.TYPEAPPLY)
            ? ((JCTypeApply) clazz).clazz
            : clazz;

        JCExpression clazzid1 = clazzid; // The same in fully qualified form

        if (tree.encl != null) {
            // We are seeing a qualified new, of the form
            //    <expr>.new C <...> (...) ...
            // In this case, we let clazz stand for the name of the
            // allocated class C prefixed with the type of the qualifier
            // expression, so that we can
            // resolve it with standard techniques later. I.e., if
            // <expr> has type T, then <expr>.new C <...> (...)
            // yields a clazz T.C.
            Type encltype = chk.checkRefType(tree.encl.pos(),
                                             attribExpr(tree.encl, env));
            clazzid1 = make.at(clazz.pos).Select(make.Type(encltype),
                                                 ((JCIdent) clazzid).name);
            if (clazz.getTag() == JCTree.TYPEAPPLY)
                clazz = make.at(tree.pos).
                    TypeApply(clazzid1,
                              ((JCTypeApply) clazz).arguments);
            else
                clazz = clazzid1;
//          System.out.println(clazz + " generated.");//DEBUG
        }

        // Attribute clazz expression and store
        // symbol + type back into the attributed tree.
        Type clazztype = chk.checkClassType(
            tree.clazz.pos(), attribType(clazz, env), true);
        final Type origClazzType = clazztype;
        Symbol origConstructor = null;

        chk.validate(clazz, localEnv);
        if (tree.encl != null) {
            // We have to work in this case to store
            // symbol + type back into the attributed tree.
            tree.clazz.type = clazztype;
            TreeInfo.setSymbol(clazzid, TreeInfo.symbol(clazzid1));
            clazzid.type = ((JCIdent) clazzid).sym.type;
            if (!clazztype.isErroneous()) {
                if (cdef != null && clazztype.tsym.isInterface()) {
                    log.error(tree.encl.pos(), "anon.class.impl.intf.no.qual.for.new");
                } else if (clazztype.tsym.isStatic()) {
                    log.error(tree.encl.pos(), "qualified.new.of.static.class", clazztype.tsym);
                }
            }
        } else if (!clazztype.tsym.isInterface() &&
                   clazztype.getEnclosingType().tag == CLASS) {
            // Check for the existence of an apropos outer instance
            rs.resolveImplicitThis(tree.pos(), env, clazztype);
        }

        // Attribute constructor arguments.
        List<Type> argtypes = attribArgs(tree.args, localEnv);
        List<Type> typeargtypes = attribTypes(tree.typeargs, localEnv);

        // If we have made no mistakes in the class type...
        if (clazztype.tag == CLASS) {
            // Enums may not be instantiated except implicitly
            if (allowEnums &&
                (clazztype.tsym.flags_field&Flags.ENUM) != 0 &&
                (env.tree.getTag() != JCTree.VARDEF ||
                 (((JCVariableDecl) env.tree).mods.flags&Flags.ENUM) == 0 ||
                 ((JCVariableDecl) env.tree).init != tree))
                log.error(tree.pos(), "enum.cant.be.instantiated");
            // Check that class is not abstract
            if (cdef == null &&
                (clazztype.tsym.flags() & (ABSTRACT | INTERFACE)) != 0) {
                log.error(tree.pos(), "abstract.cant.be.instantiated",
                          clazztype.tsym);
            } else if (cdef != null && clazztype.tsym.isInterface()) {
                // Check that no constructor arguments are given to
                // anonymous classes implementing an interface
                if (!argtypes.isEmpty())
                    log.error(tree.args.head.pos(), "anon.class.impl.intf.no.args");

                if (!typeargtypes.isEmpty())
                    log.error(tree.typeargs.head.pos(), "anon.class.impl.intf.no.typeargs");

                // Error recovery: pretend no arguments were supplied.
                argtypes = List.nil();
                typeargtypes = List.nil();
            }

            // Resolve the called constructor under the assumption
            // that we are referring to a superclass instance of the
            // current instance (JLS ???).
            else {
                localEnv.info.selectSuper = cdef != null;
                localEnv.info.varArgs = false;
                tree.constructor = rs.resolveConstructor(
                    tree.pos(), localEnv, clazztype, argtypes, typeargtypes);
                tree.constructorType = checkMethod(clazztype,
                                            tree.constructor,
                                            localEnv,
                                            tree.args,
                                            argtypes,
                                            typeargtypes,
                                            localEnv.info.varArgs);
                if (localEnv.info.varArgs)
                    assert tree.constructorType.isErroneous() || tree.varargsElement != null;
            }
            
            origConstructor = tree.constructor;

            if (cdef != null) {
                // We are seeing an anonymous class instance creation.
                // In this case, the class instance creation
                // expression
                //
                //    E.new <typeargs1>C<typargs2>(args) { ... }
                //
                // is represented internally as
                //
                //    E . new <typeargs1>C<typargs2>(args) ( class <empty-name> { ... } )  .
                //
                // This expression is then *transformed* as follows:
                //
                // (1) add a STATIC flag to the class definition
                //     if the current environment is static
                // (2) add an extends or implements clause
                // (3) add a constructor.
                //
                // For instance, if C is a class, and ET is the type of E,
                // the expression
                //
                //    E.new <typeargs1>C<typargs2>(args) { ... }
                //
                // is translated to (where X is a fresh name and typarams is the
                // parameter list of the super constructor):
                //
                //   new <typeargs1>X(<*nullchk*>E, args) where
                //     X extends C<typargs2> {
                //       <typarams> X(ET e, args) {
                //         e.<typeargs1>super(args)
                //       }
                //       ...
                //     }
                if (Resolve.isStatic(env)) cdef.mods.flags |= STATIC;

                if (clazztype.tsym.isInterface()) {
                    cdef.implementing = List.of(clazz);
                } else {
                    cdef.extending = clazz;
                }

                attribStat(cdef, localEnv);

                // If an outer instance is given,
                // prefix it to the constructor arguments
                // and delete it from the new expression
                if (tree.encl != null && !clazztype.tsym.isInterface()) {
                    tree.args = tree.args.prepend(makeNullCheck(tree.encl));
                    argtypes = argtypes.prepend(tree.encl.type);
                    tree.encl = null;
                }

                // Reassign clazztype and recompute constructor.
                clazztype = cdef.sym.type;
                Symbol sym = rs.resolveConstructor(
                    tree.pos(), localEnv, clazztype, argtypes,
                    typeargtypes, true, tree.varargsElement != null);
                assert sym.kind < AMBIGUOUS || tree.constructor.type.isErroneous();
                tree.constructor = sym;
                tree.constructorType = checkMethod(clazztype,
                                            tree.constructor,
                                            localEnv,
                                            tree.args,
                                            argtypes,
                                            typeargtypes,
                                            localEnv.info.varArgs);
            }

            if (tree.constructor != null && tree.constructor.kind == MTH)
                owntype = clazztype;            
        }
        result = check(tree, owntype, VAL, pkind, pt);
        chk.validate(tree.typeargs, localEnv);
        
        // System.out.println("newClass: owntype = "+owntype+" constructor = "+tree.constructor+
        //                    " separable? "+((tree.constructor.flags() & SEPARABLE)!=0)+
        //                    " generated_ctor? "+((tree.constructor.flags() & GENERATEDCONSTR)!=0)+
        //                    " class closer than brackets? "+isClassCloserThanBrackets()+
        //                    " env separable? "+env.isSeparable());
        if (isClassCloserThanBrackets()) {
            if (tree.constructor != null && origConstructor != null && clazztype.tag == CLASS && tree.constructor.kind == MTH &&
                (!((cdef!=null) && (origClazzType.tsym.isInterface()))) && // and not an anonymous inner class of an interface
                (!((cdef!=null) && ((origConstructor.flags() & SEPARABLE)!=0))) && // and not an anonymous inner class of a class with an separable constructor
                (!(((tree.constructor.flags() & GENERATEDCONSTR)!=0) && 
                   (checkDefaultCtorsSeparable((MethodSymbol)tree.constructor)))) && // and not a generated constructor that is separable
                (!((cdef==null) && (tree.constructor.owner.toString().equals("java.lang.Object")) &&
                   (tree.constructor.type.toString().equals("()void")) &&
                   (tree.constructor.name.toString().equals("<init>")))) && // and not Object()
                (!((cdef==null) && (tree.constructor.owner.toString().equals("java.lang.RuntimeException")) &&
                   (tree.constructor.name.toString().equals("<init>")))) && // and not RuntimeException(...)
                (!((cdef==null) && ((tree.constructor.owner.toString().equals("java.lang.Integer")) || // allow boxed primitives
                                    (tree.constructor.owner.toString().equals("java.lang.Long")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Character")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Float")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Double")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Byte")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Short")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.Boolean")) ||
                                    (tree.constructor.owner.toString().equals("java.lang.String"))) &&
                   (tree.constructor.name.toString().equals("<init>")))) && // and not <BoxedPrimitive>(...)
                (!((cdef==null) && ((tree.constructor.owner.toString().equals("edu.rice.cs.mint.runtime.ClassCode")) || // allow Mint reflection classes
                                    (tree.constructor.owner.toString().equals("edu.rice.cs.mint.runtime.FieldCode"))) &&
                   (tree.constructor.name.toString().equals("<init>"))))) { // and not <???Code>(...)

                // mgr: escape safety check
                boolean calledMethodSeparable = ((tree.constructor.flags() & SEPARABLE)!=0);
                if (!calledMethodSeparable) setClosestBracketUnsafe(tree);
                if (env.enclMethod==null) {
                    // new in an initializer block
                    if (!calledMethodSeparable) {
                        if (env.escape!=null) {
                            // error: calling a non-separable method from inside an escape in an initializer block
                            StringBuilder sb = new StringBuilder(owntype.toString());
                            sb.append('(');
                            if (argtypes!=null) {
                                boolean first = true;
                                for(Type t: argtypes) {
                                    if (first) first = false;
                                    else sb.append(',');
                                    sb.append(t);
                                }
                            }
                            sb.append(')');
                            String callee = sb.toString();
                            log.error(tree.pos(),"cannot.call.not.separable.in.escape", callee);
                        }
                        else {
                            // maybe error, depending on constructors: calling a non-separable method from an initializer block
                            env.enclClass.containsNonSeparableInitializers = true;
                        }
                    }
                }
                else {
                    if (env.isSeparable ()) {                    
                        if (!calledMethodSeparable) {
                            StringBuilder sb = new StringBuilder(owntype.toString());
                            sb.append('(');
                            if (argtypes!=null) {
                                boolean first = true;
                                for(Type t: argtypes) {
                                    if (first) first = false;
                                    else sb.append(',');
                                    sb.append(t);
                                }
                            }
                            sb.append(')');
                            String callee = sb.toString();
                            if (env.escape==null) {
                                // within separable method
                                String caller = env.enclMethod.name.toString();
                                if (env.enclMethod.name==names.init) {
                                    sb.setLength(0);
                                    sb.append(env.enclClass.sym.toString());
                                    sb.append('(');
                                    boolean first = true;
                                    for(JCVariableDecl decl: env.enclMethod.params) {
                                        if (first) first = false;
                                        else sb.append(',');
                                        sb.append(decl.vartype);
                                    }
                                    sb.append(')');
                                    caller = sb.toString();
                                }
                                log.error(tree.pos(),"not.separable.called.from.separable", callee, caller);
                            }
                            else {
                                // within escape
                                log.error(tree.pos(),"cannot.call.not.separable.in.escape", callee);
                            }
                        }
                    }
                }
            }
        }
    }
    protected boolean checkDefaultCtorsSeparable(MethodSymbol ctor) {
        // this is a generated constructor
        // check that all the default constructors in the super classes are separable
        // System.out.println("ctor: "+ctor+" owner: "+ctor.owner+" "+ctor.owner.getClass().getSimpleName());
        ClassSymbol clazz = (ClassSymbol)ctor.owner;
        Type stype = clazz.getSuperclass();
        while(stype!=Type.noType) {
            // System.out.println("\ttype: "+stype);
            if ((stype==syms.objectType) ||
                (stype==syms.runtimeExceptionType)) return true; // we have made it all the way to Object or RuntimeException, this is separable
            Scope s = stype.tsym.members();
            boolean found = false;
            for(Symbol member: s.getElements()) {
//                System.out.println("\t\t"+member);
                if (member.isConstructor()) {
                    MethodSymbol sctor = (MethodSymbol)member;
//                    System.out.println("\t\t\tctor: "+sctor+" params: "+sctor.params+" names: "+sctor.savedParameterNames+" type: "+sctor.type);
                    if (sctor.type.asMethodType().argtypes.isEmpty()) {
                        // this is the default constructor
                        // make sure it is either generated or separable
                        boolean generated = ((sctor.flags() & GENERATEDCONSTR)!=0);
                        boolean isSeparable = ((sctor.flags() & SEPARABLE)!=0);
//                        System.out.println("\t\t\t\tdefault, generated? "+generated+", separable? "+isSeparable);
                        if (isSeparable) return true; // if this has been declared separable, then we can stop here
                        if (!generated) return false; // this is not generated, i.e. it is user-written and NOT separable, we can stop here
                        // otherwise we need to continue with the superclass
                        clazz = (ClassSymbol)sctor.owner;
                        stype = clazz.getSuperclass();
                        if ((stype==syms.objectType) ||
                            (stype==syms.runtimeExceptionType)) return true; // we have made it all the way to Object or RuntimeException, this is separable
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                // didn't find a default ctor!
                // throw new AssertionError("Could not find default constructor for "+stype);
                return false;
            }
        }
        return true;
    }
    
    /** Make an attributed null check tree.
      */
    public JCExpression makeNullCheck(JCExpression arg) {
        // optimization: X.this is never null; skip null check
        Name name = TreeInfo.name(arg);
        if (name == names._this || name == names._super) return arg;

        int optag = JCTree.NULLCHK;
        JCUnary tree = make.at(arg.pos).Unary(optag, arg);
        tree.operator = syms.nullcheck;
        tree.type = arg.type;
        return tree;
    }

    public void visitNewArray(JCNewArray tree) {
        Type owntype = types.createErrorType(tree.type);
        Type elemtype;
        if (tree.elemtype != null) {
            elemtype = attribType(tree.elemtype, env);
            chk.validate(tree.elemtype, env);
            owntype = elemtype;
            for (List<JCExpression> l = tree.dims; l.nonEmpty(); l = l.tail) {
                attribExpr(l.head, env, syms.intType);
                owntype = new ArrayType(owntype, syms.arrayClass);
            }
        } else {
            // we are seeing an untyped aggregate { ... }
            // this is allowed only if the prototype is an array
            if (pt.tag == ARRAY) {
                elemtype = types.elemtype(pt);
            } else {
                if (pt.tag != ERROR) {
                    log.error(tree.pos(), "illegal.initializer.for.type",
                              pt);
                }
                elemtype = types.createErrorType(pt);
            }
        }
        if (tree.elems != null) {
            attribExprs(tree.elems, env, elemtype);
            owntype = new ArrayType(elemtype, syms.arrayClass);
        }
        if (!types.isReifiable(elemtype))
            log.error(tree.pos(), "generic.array.creation");
        result = check(tree, owntype, VAL, pkind, pt);
    }

    public void visitParens(JCParens tree) {
        Type owntype = attribTree(tree.expr, env, pkind, pt);
        result = check(tree, owntype, pkind, pkind, pt);
        Symbol sym = TreeInfo.symbol(tree);
        if (sym != null && (sym.kind&(TYP|PCK)) != 0)
            log.error(tree.pos(), "illegal.start.of.type");
    }

    public void visitAssign(JCAssign tree) {
        Type owntype = attribTree(tree.lhs, env.dup(tree), VAR, Type.noType);
        Type capturedType = capture(owntype);
        attribExpr(tree.rhs, env, owntype);
        result = check(tree, capturedType, VAL, pkind, pt);
        
        // mgr: escape safety check
        checkAssignSeparable(tree.lhs);
    }
    
    // Definition 1: A type is code-free if
    // (1) it is not a subtype of Code<T>,
    // (2) the types of all of its fields are code-free
    // (3) all of its methods' parameter and return types are code-free, and
    // (4) its class is final.
    protected boolean checkTypeCodeFree(Type t) {
        if (options.get("-verbose") != null) System.out.println("checkTypeCodeFree "+t);
        if (t.tsym.isSubClass(syms.codeFreeType.tsym, types)) {
            if (options.get("-verbose") != null) System.out.println("\tsubclass of CodeFree --> code-free");
            return true;
        }
        if (isPrimitiveOrBoxedOrArrayThereof(t)) {
            if (options.get("-verbose") != null) System.out.println("\tprimitive, boxed, String or array thereof --> code-free");
            return true;
        }
        if ((t.tsym.flags_field & FINAL)==0) { // (4)
            if (options.get("-verbose") != null) System.out.println("\tnot final --> not code-free");
            return false;
        }
        return checkTypeCodeFreeHelper(t);
    }

    private boolean checkTypeCodeFreeHelper(Type stype) {
        if (options.get("-verbose") != null) System.out.println("checkTypeCodeFreeHelper "+stype);
        if (options.get("-verbose") != null) System.out.println("\ttype: "+stype);
        if (isCodeType(stype)) { // (1)
            if (options.get("-verbose") != null) System.out.println("\tcode type: "+stype+" --> not code-free");
            return false;
        }
        if (stype.tsym.isSubClass(syms.codeFreeType.tsym, types)) {
            if (options.get("-verbose") != null) System.out.println("\tsubclass of CodeFree --> code-free");
            return true;
        }
        
        if (stype==syms.objectType) {
            if (options.get("-verbose") != null) System.out.println("\treached Object --> code-free");
            return true; // we have made it all the way to Object, this is separable
        }
        Scope s = stype.tsym.members();
        for(Symbol member: s.getElements()) {
            if (member.kind==VAR) {
                VarSymbol v = (VarSymbol)member;
                if (v.getKind()==ElementKind.FIELD) {
                    if (options.get("-verbose") != null) System.out.println("\tfield "+v+", type "+v.type);
                    if (!checkTypeCodeFree(v.type)) {
                        if (options.get("-verbose") != null) System.out.println("\tfield "+v+", type "+v.type+" not code-free --> not code-free");
                        return false;
                    }
                }
            }
            else if (member.kind==MTH) {
                MethodSymbol m = (MethodSymbol)member;
                if (options.get("-verbose") != null) System.out.println("\tmethod "+m+", return type "+m.getReturnType());
                if ((m.getReturnType()!=syms.voidType) &&
                    (m.getReturnType()!=syms.javaLangVoidType) &&
                    (!checkTypeCodeFree(m.getReturnType()))) {
                    if (options.get("-verbose") != null) System.out.println("\tmethod "+m+", return type "+m.getReturnType()+" not code-free --> not code-free");
                    return false;
                }
                if (m.params!=null) {
                    // also check method parameters
                    for(VarSymbol mParam: m.params) {
                        if (options.get("-verbose") != null) System.out.println("\tmethod "+m+", parameter type "+mParam.type);
                        if (mParam.type!=syms.javaLangVoidType &&
                            (!checkTypeCodeFree(mParam.type))) {
                            if (options.get("-verbose") != null) System.out.println("\tmethod "+m+", parameter type "+mParam.type+" not code-free --> not code-free");
                            return false;
                        }
                    }
                }
            }
        }
        
        if (stype.tsym instanceof ClassSymbol) {
            ClassSymbol csym = (ClassSymbol)stype.tsym;
            if (options.get("-verbose") != null) System.out.println("\tsuperclass of "+stype+", type "+csym.getSuperclass());
            if (!checkTypeCodeFreeHelper(csym.getSuperclass())) {
                if (options.get("-verbose") != null) System.out.println("\tsuperclass of "+stype+", type "+csym.getSuperclass()+" not code-free --> not code-free");
                return false;
            }
            
            for(Type iface: csym.getInterfaces()) {
                if (options.get("-verbose") != null) System.out.println("\tinterfacfe of "+stype+", type "+iface);
                if (!checkTypeCodeFreeHelper(iface)) {
                    if (options.get("-verbose") != null) System.out.println("\tsuperclass of "+stype+", type "+iface+" not code-free --> not code-free");
                    return false;
                }
            }
        }
        return true;
    }
    
    protected void checkAssignSeparable(JCExpression lhs) {
        if (lhs instanceof JCParens) {
            checkAssignSeparable(((JCParens)lhs).expr);
            return;
        }
        if (options.get("-verbose") != null) System.out.println("checkAssignSeparable: "+lhs);
        if (env.enclMethod==null) {
            if (options.get("-verbose") != null) System.out.println("\tenv.enclMethod==null --> initializer");
            if (lhs instanceof JCIdent) {
                JCIdent ident = (JCIdent)lhs;
                switch (ident.sym.getKind()) {
                    case FIELD: {
                        // separability for `this' field accesses in initializers
                        if (!checkTypeCodeFree(ident.sym.type)) {
                            env.enclClass.containsNonSeparableInitializers = true;
                        }
                        break;
                    }
                    case PARAMETER:
                    case LOCAL_VARIABLE: {
                        // separability for parameters and local variables in initializers
                        // allow parameters and local variables
                        // if bound inside this escape
                        if (env.escape!=null) {
                            Env<AttrContext> defineEnv = env;
                            Scope defineScope = defineEnv.info.scope.lookup(ident.sym.name).scope;
                            while(defineScope==null && defineEnv.outer!=null) {
                                defineEnv = defineEnv.next;
                                defineScope = defineEnv.info.scope.lookup(ident.sym.name).scope;
                            }
                            if (defineEnv.escape!=env.escape) {
                                if (!checkTypeCodeFree(ident.sym.type)) {
                                    env.enclClass.containsNonSeparableInitializers = true;
                                }
                            }
                        }
                        else {
                            // must be inside an separable method
                            // all parameters and local variables are okay
                            // System.out.println("\tno escape");
                        }
                        break;
                    }
                    case PACKAGE:
                    case ENUM:
                    case CLASS:
                    case ANNOTATION_TYPE:
                    case INTERFACE:
                    case ENUM_CONSTANT:
                    case EXCEPTION_PARAMETER:
                    case METHOD:
                    case CONSTRUCTOR:
                    case STATIC_INIT:
                    case INSTANCE_INIT:
                    case TYPE_PARAMETER:
                    case OTHER:
                    default: // unexpected, don't do anything
                        break;
                }
            }
            else if (lhs instanceof JCFieldAccess) {
                // separability for field accesses in initializers
                JCFieldAccess fa = (JCFieldAccess)lhs;
                if (!checkTypeCodeFree(fa.sym.type)) {
                    env.enclClass.containsNonSeparableInitializers = true;
                }
            }

            else if (lhs instanceof JCArrayAccess) {
                // separability for array accesses in initializers
                JCArrayAccess aa = (JCArrayAccess)lhs;
                if (!checkTypeCodeFree(aa.type)) {
                    env.enclClass.containsNonSeparableInitializers = true;
                }
            }
            else {
                throw new AssertionError("Expected a JCIdent, JCFieldAccess, JCArrayAccess, or JCParens as lhs of an assignment, was "+
                                         lhs.getClass().getSimpleName()+" in "+lhs);
            }
        }
        else {
            if (options.get("-verbose") != null) System.out.println("\tenv.enclMethod!=null --> method");
            if (env.isSeparable ()) {
                if (lhs instanceof JCIdent) {
                    JCIdent ident = (JCIdent)lhs;
                    // System.out.println("\tsym = "+ident.sym+" kind = "+ident.sym.getKind());
                    switch (ident.sym.getKind()) {
                        case FIELD: {
                            // separability for `this' field accesses in methods
                            // NOTE: exception for assignment to not code-free field while in constructor!
                            if ((env.enclMethod.name!=names.init) && !checkTypeCodeFree(ident.sym.type)) {
                                // assignment to not code-free field in a separable method that is not the constructor
                                log.error(lhs.pos(), "assignment.to.non.codefree.fields.in.separable");
                            }
                            break;
                        }
                        case PARAMETER:
                        case LOCAL_VARIABLE: {
                            // separability for parameters and local variables in methods
                            // allow parameters and local variables
                            // if bound inside this escape
                            if (env.escape!=null) {
                                Env<AttrContext> defineEnv = env;
                                Scope defineScope = defineEnv.info.scope.lookup(ident.sym.name).scope;
                                while(defineScope==null && defineEnv.outer!=null) {
                                    defineEnv = defineEnv.next;
                                    defineScope = defineEnv.info.scope.lookup(ident.sym.name).scope;
                                }
                                if (defineEnv.escape!=env.escape) {
                                    if (!checkTypeCodeFree(ident.sym.type)) {
                                        log.error(lhs.pos(), "assignment.in.escape.to.non.codefree.local.outside.escape");
                                    }
                                }
                            }
                            else {
                                // must be inside an separable method
                                // all parameters and local variables are okay
                                // System.out.println("\tno escape");
                            }
                            break;
                        }
                        case PACKAGE:
                        case ENUM:
                        case CLASS:
                        case ANNOTATION_TYPE:
                        case INTERFACE:
                        case ENUM_CONSTANT:
                        case EXCEPTION_PARAMETER:
                        case METHOD:
                        case CONSTRUCTOR:
                        case STATIC_INIT:
                        case INSTANCE_INIT:
                        case TYPE_PARAMETER:
                        case OTHER:
                        default: // unexpected, don't do anything
                            break;
                    }
                }
                else if (lhs instanceof JCFieldAccess) {
                    // separability for field accesses in methods
                    JCFieldAccess fa = (JCFieldAccess)lhs;
                    if ((fa.selected instanceof JCIdent) && (((JCIdent)fa.selected).name==names._this)) {
                        // this.x access
                        // NOTE: exception for assignment to not code-free field while in constructor!
                        if ((env.enclMethod.name!=names.init) && !checkTypeCodeFree(fa.sym.type)) {
                            // assignment to not code-free field in a separable method that is not the constructor
                            log.error(lhs.pos(), "assignment.to.non.codefree.fields.in.separable");
                        }
                    }
                    else {
                        // not a this.x access
                        if (!checkTypeCodeFree(fa.sym.type)) {
                            log.error(lhs.pos(), "assignment.to.non.codefree.fields.in.separable");
                        }
                    }
                }
                else if (lhs instanceof JCArrayAccess) {
                    // separability for array accesses in methods
                    JCArrayAccess aa = (JCArrayAccess)lhs;
                    if (!checkTypeCodeFree(aa.type)) {
                        log.error(lhs.pos(), "assignment.to.non.codefree.arrays.in.separable");
                    }
                }
                else {
                    throw new AssertionError("Expected a JCIdent, JCFieldAccess, JCArrayAccess, or JCParens as lhs of an assignment, was "+
                                             lhs.getClass().getSimpleName()+" in "+lhs);
                }
                // System.out.println("In separable/escape: "+lhs+" "+lhs.getClass().getSimpleName());
            }
        }
    }

    public void visitAssignop(JCAssignOp tree) {
        // Attribute arguments.
        Type owntype = attribTree(tree.lhs, env, VAR, Type.noType);
        Type operand = attribExpr(tree.rhs, env);
        // Find operator.
        Symbol operator = tree.operator = rs.resolveBinaryOperator(
            tree.pos(), tree.getTag() - JCTree.ASGOffset, env,
            owntype, operand);

        if (operator.kind == MTH) {
            chk.checkOperator(tree.pos(),
                              (OperatorSymbol)operator,
                              tree.getTag() - JCTree.ASGOffset,
                              owntype,
                              operand);
            chk.checkDivZero(tree.rhs.pos(), operator, operand);
            chk.checkCastable(tree.rhs.pos(),
                              operator.type.getReturnType(),
                              owntype);
        }
        result = check(tree, owntype, VAL, pkind, pt);

        // mgr: escape safety check
        checkAssignSeparable(tree.lhs);
    }

    public void visitUnary(JCUnary tree) {
        // Attribute arguments.
        Type argtype = (JCTree.PREINC <= tree.getTag() && tree.getTag() <= JCTree.POSTDEC)
            ? attribTree(tree.arg, env, VAR, Type.noType)
            : chk.checkNonVoid(tree.arg.pos(), attribExpr(tree.arg, env));

        // Find operator.
        Symbol operator = tree.operator =
            rs.resolveUnaryOperator(tree.pos(), tree.getTag(), env, argtype);

        Type owntype = types.createErrorType(tree.type);
        if (operator.kind == MTH) {
            owntype = (JCTree.PREINC <= tree.getTag() && tree.getTag() <= JCTree.POSTDEC)
                ? tree.arg.type
                : operator.type.getReturnType();
            int opc = ((OperatorSymbol)operator).opcode;

            // If the argument is constant, fold it.
            if (argtype.constValue() != null) {
                Type ctype = cfolder.fold1(opc, argtype);
                if (ctype != null) {
                    owntype = cfolder.coerce(ctype, owntype);

                    // Remove constant types from arguments to
                    // conserve space. The parser will fold concatenations
                    // of string literals; the code here also
                    // gets rid of intermediate results when some of the
                    // operands are constant identifiers.
                    if (tree.arg.type.tsym == syms.stringType.tsym) {
                        tree.arg.type = syms.stringType;
                    }
                }
            }
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    public void visitBinary(JCBinary tree) {
        // Attribute arguments.
        Type left = chk.checkNonVoid(tree.lhs.pos(), attribExpr(tree.lhs, env));
        Type right = chk.checkNonVoid(tree.lhs.pos(), attribExpr(tree.rhs, env));

        // Find operator.
        Symbol operator = tree.operator =
            rs.resolveBinaryOperator(tree.pos(), tree.getTag(), env, left, right);

        Type owntype = types.createErrorType(tree.type);
        if (operator.kind == MTH) {
            owntype = operator.type.getReturnType();
            int opc = chk.checkOperator(tree.lhs.pos(),
                                        (OperatorSymbol)operator,
                                        tree.getTag(),
                                        left,
                                        right);

            // If both arguments are constants, fold them.
            if (left.constValue() != null && right.constValue() != null) {
                Type ctype = cfolder.fold2(opc, left, right);
                if (ctype != null) {
                    owntype = cfolder.coerce(ctype, owntype);

                    // Remove constant types from arguments to
                    // conserve space. The parser will fold concatenations
                    // of string literals; the code here also
                    // gets rid of intermediate results when some of the
                    // operands are constant identifiers.
                    if (tree.lhs.type.tsym == syms.stringType.tsym) {
                        tree.lhs.type = syms.stringType;
                    }
                    if (tree.rhs.type.tsym == syms.stringType.tsym) {
                        tree.rhs.type = syms.stringType;
                    }
                }
            }

            // Check that argument types of a reference ==, != are
            // castable to each other, (JLS???).
            if ((opc == ByteCodes.if_acmpeq || opc == ByteCodes.if_acmpne)) {
                if (!types.isCastable(left, right, new Warner(tree.pos()))) {
                    log.error(tree.pos(), "incomparable.types", left, right);
                }
            }

            chk.checkDivZero(tree.rhs.pos(), operator, right);
        }
        result = check(tree, owntype, VAL, pkind, pt);
    }

    public void visitTypeCast(JCTypeCast tree) {
        Type clazztype = attribType(tree.clazz, env);
        chk.validate(tree.clazz, env);
        Type exprtype = attribExpr(tree.expr, env, Infer.anyPoly);
        Type owntype = chk.checkCastable(tree.expr.pos(), exprtype, clazztype);
        if (exprtype.constValue() != null)
            owntype = cfolder.coerce(exprtype, owntype);
        result = check(tree, capture(owntype), VAL, pkind, pt);

        // mgr: disallow casts to Code
        // NOTE: we don't do that anymore
        // boolean isCastToCode = false;
        // if (tree.clazz instanceof JCIdent) {
        //     JCIdent clazz = (JCIdent)tree.clazz;
        //     Type casttype = ((ClassSymbol)clazz.sym).type;
        //     isCastToCode = types.isSubtypeUnchecked(casttype,syms.codeType);
        // }
        // else if (tree.clazz instanceof JCTypeApply) {
        //     JCTypeApply tapp = (JCTypeApply)tree.clazz;
        //     JCIdent clazz = (JCIdent)tapp.clazz;
        //     Type casttype = ((ClassSymbol)clazz.sym).type;
        //     isCastToCode = types.isSubtypeUnchecked(casttype,syms.codeType);
        // }
        // if (isCastToCode) {
        //     log.error(tree.pos(), "may.not.cast.to.code");
        // }
    }

    public void visitTypeTest(JCInstanceOf tree) {
        Type exprtype = chk.checkNullOrRefType(
            tree.expr.pos(), attribExpr(tree.expr, env));
        Type clazztype = chk.checkReifiableReferenceType(
            tree.clazz.pos(), attribType(tree.clazz, env));
        chk.validate(tree.clazz, env);
        chk.checkCastable(tree.expr.pos(), exprtype, clazztype);
        result = check(tree, syms.booleanType, VAL, pkind, pt);
    }

    public void visitIndexed(JCArrayAccess tree) {
        Type owntype = types.createErrorType(tree.type);
        Type atype = attribExpr(tree.indexed, env);
        attribExpr(tree.index, env, syms.intType);
        if (types.isArray(atype))
            owntype = types.elemtype(atype);
        else if (atype.tag != ERROR)
            log.error(tree.pos(), "array.req.but.found", atype);
        if ((pkind & VAR) == 0) owntype = capture(owntype);
        result = check(tree, owntype, VAR, pkind, pt);
    }

    public void visitIdent(JCIdent tree) {
        Symbol sym;
        boolean varArgs = false;
        
        // Find symbol
        if (pt.tag == METHOD || pt.tag == FORALL) {
            // If we are looking for a method, the prototype `pt' will be a
            // method type with the type of the call's arguments as parameters.
            env.info.varArgs = false;
            sym = rs.resolveMethod(tree.pos(), env, tree.name, pt.getParameterTypes(), pt.getTypeArguments());
            varArgs = env.info.varArgs;
        } else if (tree.sym != null && tree.sym.kind != VAR) {
            sym = tree.sym;
        } else {
            sym = rs.resolveIdent(tree.pos(), env, tree.name, pkind);
        }
        tree.sym = sym;

        // (1) Also find the environment current for the class where
        //     sym is defined (`symEnv').
        // Only for pre-tiger versions (1.4 and earlier):
        // (2) Also determine whether we access symbol out of an anonymous
        //     class in a this or super call.  This is illegal for instance
        //     members since such classes don't carry a this$n link.
        //     (`noOuterThisPath').
        Env<AttrContext> symEnv = env;
        boolean noOuterThisPath = false;
        if (env.enclClass.sym.owner.kind != PCK && // we are in an inner class
            (sym.kind & (VAR | MTH | TYP)) != 0 &&
            sym.owner.kind == TYP &&
            tree.name != names._this && tree.name != names._super) {

            // Find environment in which identifier is defined.
            while (symEnv.outer != null &&
                   !sym.isMemberOf(symEnv.enclClass.sym, types)) {
                if ((symEnv.enclClass.sym.flags() & NOOUTERTHIS) != 0)
                    noOuterThisPath = !allowAnonOuterThis;
                symEnv = symEnv.outer;
            }
        }

        // If symbol is a variable, ...
        if (sym.kind == VAR) {
            VarSymbol v = (VarSymbol)sym;

            // ..., evaluate its initializer, if it has one, and check for
            // illegal forward reference.
            checkInit(tree, env, v, false);

            // If symbol is a local variable accessed from an embedded
            // inner class check that it is final.
            if (v.owner.kind == MTH &&
                v.owner != env.info.scope.owner &&
                (v.flags_field & FINAL) == 0) {
                log.error(tree.pos(),
                          "local.var.accessed.from.icls.needs.final",
                          v);
            }

            // If we are expecting a variable (as opposed to a value), check
            // that the variable is assignable in the current environment.
            if (pkind == VAR)
                checkAssignable(tree.pos(), v, null, env);
            
            // mgr: CSP level check
            // Find the scope in which the variable was defined
            Env<AttrContext> defineEnv = env;
            Scope defineScope = defineEnv.info.scope.lookup(sym.name).scope;
            while(defineScope==null && defineEnv.outer!=null) {
                defineEnv = defineEnv.next;
                defineScope = defineEnv.info.scope.lookup(sym.name).scope;
            }

            // Ensure that the level in which the variable was defined isn't greater than the current level
            // System.out.println("sym "+sym+" level="+env.getLevel()+" define level="+defineEnv.getLevel());
            // System.out.println("\tis member of "+env.enclClass.sym+"? "+sym.isMemberOf(env.enclClass.sym, types));
            if (sym.isMemberOf(env.enclClass.sym, types)) {
                // this symbol is a member of this class, don't CSP it
                tree.cspCount = 0;
            }
            else {
                tree.cspCount = env.getLevel()-defineEnv.getLevel();
            }
            if (tree.cspCount<0) {
                log.error(tree.pos(), "escape.level", defineEnv.getLevel(), env.getLevel());
                return;
            }
            else if (tree.cspCount>0) {
                cspVariableCheck(sym, tree);
            }
            if (env.enclBrackets.size()>0) {
                // find the environment where the bracket was created
                Env<AttrContext> bracketEnclEnv = null;
                Env<AttrContext> e = env;
                while(e!=null) {
                    if (e.tree==env.enclBrackets.head) {
                        bracketEnclEnv = e;
                        break;
                    }
                    e = e.next;
                }

                // local variables and parameters outside the brackets must be final
/* FIXME: isn't this covered in cspVariableCheck?
                if (((tree.sym.getKind()==ElementKind.PARAMETER) || (tree.sym.getKind()==ElementKind.LOCAL_VARIABLE)) &&
                    ((tree.sym.flags() & FINAL)==0)) {                    
                    // determine if the binding occurrence was outside the brackets
                    // by starting outside the bracket's env and searching for defineEnv
                    e = bracketEnclEnv.next;
                    while(e!=null) {
                        if (e==defineEnv) {
                            log.error(tree.pos(), "local.var.accessed.from.brackets.needs.final", sym);
                            break;
                        }
                        e = e.next;
                    }
                }
*/

                // local variables and parameters outside the brackets must be static
                // FIXME: shouldn't this also go in cspVariableCheck?
                if (((tree.sym.getKind()== edu.rice.cs.mint.comp.javax.lang.model.element.ElementKind.FIELD) || (tree.sym.getKind()==ElementKind.METHOD)) &&
                    ((tree.sym.flags() & STATIC)==0)) {
                    // System.out.println("accessing non-static field or method, "+tree.sym.owner);
                    // System.out.println(env);
                    // determine if the binding occurrence was outside the brackets
                    e = env;
                    while(e!=null) {
                        // System.out.println("Checking "+e.info.scope.owner+" "+e.info.scope.owner.getClass().getSimpleName());
                        if (e==bracketEnclEnv) {
                            log.error(tree.pos(), "may.not.directly.access.enclosing", tree.sym.name.toString());
                            break;
                        }
                        if ((e.info.scope.owner.kind & TYP) != 0) {
                            // System.out.println("\tis "+e.info.scope.owner.type+" subtype of "+tree.sym.owner.type+"?");
                            // this is a class, check if its a subclass of tree.sym.owner
                            if ((e.info.scope.owner.type.toString().equals(tree.sym.owner.type.toString())) ||
                                (types.isSubtype(e.info.scope.owner.type,tree.sym.owner.type))) {
                                // System.out.println("yes");
                                break;
                            }
                        }
                        e = e.next;
                    }
                }
                
                // prevent use of this inside brackets
                if (tree.sym.name==names._this) {
                    // find out tree.sym.owner is an enclosing class of bracketEnclEnv.enclClass.sym.type
                    Type t = bracketEnclEnv.enclClass.sym.type;
                    while(t!=null) {
                        if (tree.sym.owner.type.toString().equals(t.toString())) {
                            // tree.sym.owner is an enclosing class of bracketEnclClass
                            // that means that the this belongs to a class outside the brackets
                            log.error(tree.pos(), "may.not.access.enclosing.this");
                            break;
                        }
                        t = t.getEnclosingType();
                    }
                }
            }
        }

        // In a constructor body,
        // if symbol is a field or instance method, check that it is
        // not accessed before the supertype constructor is called.
        if ((symEnv.info.isSelfCall || noOuterThisPath) &&
            (sym.kind & (VAR | MTH)) != 0 &&
            sym.owner.kind == TYP &&
            (sym.flags() & STATIC) == 0) {
            chk.earlyRefError(tree.pos(), sym.kind == VAR ? sym : thisSym(tree.pos(), env));
        }
        Env<AttrContext> env1 = env;
        if (sym.kind != ERR && sym.kind != TYP && sym.owner != null && sym.owner != env1.enclClass.sym) {
            // If the found symbol is inaccessible, then it is
            // accessed through an enclosing instance.  Locate this
            // enclosing instance:
            while (env1.outer != null && !rs.isAccessible(env, env1.enclClass.sym.type, sym))
                env1 = env1.outer;
        }
        result = checkId(tree, env1.enclClass.sym.type, sym, env, pkind, pt, varArgs);
    }
    
    /**
     * Cache for queries to hasCodeFields.
     * Key: type, value: true if the key has code fields
     */
    protected HashMap<String,Boolean> hasCodeFieldsCache = new HashMap<String,Boolean>(); 
    
    /**
     * Return true if the type has code fields. Uses the cache.
     * 
     * @param type type to check
     * @return true if the type has code fields
     */
    protected boolean hasCodeFields(Type type) {
        Boolean result = hasCodeFieldsCache.get(type.toString());
        if (result==null) {
            if (type!=syms.objectType) hasCodeFields(syms.objectType);
                
            result = hasCodeFields(type, new HashSet<String>());
            hasCodeFieldsCache.put(type.toString(), result);
            // System.out.println("hasCodeFields: "+type.toString()+" "+result);
        }
        return result;
    }
    
    /**
     * Return true if the type has code fields. Uses the passed
     * hash set to prevent infinite recursion in cycles.
     * 
     * @param type type to check
     * @param visited hash set of types already visited
     * @return true if the type has code fields
     */
    protected boolean hasCodeFields(Type type, HashSet<String> visited) {
        if (visited.contains(type.toString())) return false; // this type has already been visited, short out
        
        ListBuffer<Type> checked = new ListBuffer<Type>();
        Boolean result = false;
        try {
            if (isPrimitiveOrBoxedOrArrayThereof(type)) return false;
            if (type==syms.stringType) return false; // allowing String (it has a field that returns hasCodeFields==true, but it is final)
            if (type.tag==ARRAY) return hasCodeFields(((ArrayType)type).elemtype, visited);
            List<Type> from = List.<Type>nil();
            List<Type> to = List.nil();
            boolean isRaw;
            do {
                if (type.tag==CLASS) {
                    if (type.toString().indexOf('<')==-1) checked.add(type); // if this class does not contain generics
                    if (visited.contains(type.toString())) { result = false; return result; }
                    result = hasCodeFieldsCache.get(type.toString());
                    if (result!=null) { return result; }
                    visited.add(type.toString()); // now this type has been visited
                    
                    ClassType t = (ClassType)type;
                    Scope s = t.tsym.members();
                    isRaw = t.isRaw();
//                    System.out.print("\n\n========================================\nCLASS "+t+"\n\tparams = "+t.typarams_field+"\n\tvars   =");
                    ListBuffer<Type> lbVars = new ListBuffer<Type>();
                    ListBuffer<Type> lbArgs = new ListBuffer<Type>();
                    List<Type> typarams_field = t.typarams_field;
                    for(TypeSymbol tv: t.tsym.getTypeParameters()) {
//                        System.out.print(" "+tv.type);
                        lbVars.add(tv.type);
                        if (isRaw) {
                            lbArgs.add(syms.objectType);
                        }
                        else {
                            Type head = typarams_field.head;
                            if (head.isSuperBound() && (!head.isUnbound())) head = syms.objectType;
                            lbArgs.add(head);
                            typarams_field = typarams_field.tail;
                        }
                    }
//                    System.out.println("\n\traw? "+isRaw);
                    from = lbVars.toList().appendList(from);
                    to = lbArgs.toList().appendList(to);
                    t = (ClassType)substList(t, from, to);
                    for(Symbol member: s.getElements()) {
                        if (((member.flags()&PRIVATE)==0)&&(member.getKind()==ElementKind.FIELD)) {
//                            Type memberType = types.memberType(t, member);
                            Type memberType = substList(member.type, from, to);
                            memberType = substList(memberType, from, to);
                            memberType = types.upperBound(memberType);
//                            System.out.println(t+" "+member+" ----> "+memberType);
                            
//                            System.out.println("\tisObject? "+(memberType==syms.objectType)+" isCode? "+isCodeType(memberType));
                            // TODO: Do we want to allow Object? This may be good enough if we disallow casts to Code.
                            // if ((memberType==syms.objectType) || (isCodeType(memberType))) { System.out.println("! Object or Code"); result = true; return result; }
                            if (!memberType.tsym.isSubClass(syms.codeFreeType.tsym, types)) {
                                if (isCodeType(memberType)) { 
                                    // System.out.println("    ### isCodeType");
                                    result = true;
                                    return result;
                                }
                                if ((memberType.tag==CLASS) && ((memberType.tsym.flags() & FINAL)==0)) {
                                    // not final class
                                    // System.out.println("    ### not final");
                                    result = true;
                                    return result;
                                }
                                if ((hasCodeMethods(memberType, visited) || hasCodeFields(memberType))) {
                                    // System.out.println("    ### hasCodeMethods");
                                    result = true;
                                    return result;
                                }
                            }
                        }
                    }
                    type = ((ClassSymbol)t.tsym).getSuperclass();
//                    System.out.println("superclass: "+type);
                }
                else {
                    type = Type.noType;
                }
            } while(type!=Type.noType);
//            System.out.println("done.");
            result = false;
            return result;
        }
        finally {
            // if there are no generics involved, then we can already cache the classes
            for(Type c: checked.toList()) {
                hasCodeFieldsCache.put(c.toString(), result);
//                System.out.println("checked: "+c.toString()+" "+result);
            }
        }
    }
    
    
    /**
     * Cache for queries to hasCodeMethods.
     * Key: type, value: true if the key has code methods
     */
    protected HashMap<String,Boolean> hasCodeMethodsCache = new HashMap<String,Boolean>(); 
    
    /**
     * Return true if the type has code methods. Uses the cache.
     * 
     * @param type type to check
     * @return true if the type has code methods
     */
    protected boolean hasCodeMethods(Type type) {
        Boolean result = hasCodeMethodsCache.get(type.toString());
        if (result==null) {
            if (type!=syms.objectType) hasCodeMethods(syms.objectType);
                
            result = hasCodeMethods(type, new HashSet<String>());
            hasCodeMethodsCache.put(type.toString(), result);
            // System.out.println("hasCodeMethods: "+type.toString()+" "+result);
        }
        return result;
    }
    
    /**
     * Return true if the type has code methods. Uses the passed
     * hash set to prevent infinite recursion in cycles.
     * 
     * @param type type to check
     * @param visited hash set of types already visited
     * @return true if the type has code methods
     */
    protected boolean hasCodeMethods(Type type, HashSet<String> visited) {
        if (visited.contains(type.toString())) return false; // this type has already been visited, short out
        
        ListBuffer<Type> checked = new ListBuffer<Type>();
        Boolean result = false;
        try {
            if (isPrimitiveOrBoxedOrArrayThereof(type)) return false;
            if (type==syms.stringType) return false; // allowing String (it has a method that returns hasCodeMethods==true, but it is final)
            if (type.tag==ARRAY) return hasCodeMethods(((ArrayType)type).elemtype, visited);
            List<Type> from = List.<Type>nil();
            List<Type> to = List.nil();
            boolean isRaw;
            do {
                if (type.tag==CLASS) {
                    if (type.toString().indexOf('<')==-1) checked.add(type); // if this class does not contain generics
                    if (visited.contains(type.toString())) { result = false; return result; }
                    result = hasCodeMethodsCache.get(type.toString());
                    if (result!=null) { return result; }
                    visited.add(type.toString()); // now this type has been visited
                    
                    ClassType t = (ClassType)type;
                    Scope s = t.tsym.members();
                    isRaw = t.isRaw();
                    // System.out.print("\n\n========================================\nCLASS "+t+"\n\tparams = "+t.typarams_field+"\n\tvars   =");
                    ListBuffer<Type> lbVars = new ListBuffer<Type>();
                    ListBuffer<Type> lbArgs = new ListBuffer<Type>();
                    List<Type> typarams_field = t.typarams_field;
                    for(TypeSymbol tv: t.tsym.getTypeParameters()) {
                        // System.out.print(" "+tv.type);
                        lbVars.add(tv.type);
                        if (isRaw) {
                            lbArgs.add(syms.objectType);
                        }
                        else {
                            Type head = typarams_field.head;
                            if (head.isSuperBound() && (!head.isUnbound())) head = syms.objectType;
                            lbArgs.add(head);
                            typarams_field = typarams_field.tail;
                        }
                    }
                    // System.out.println("\n\traw? "+isRaw);
                    from = lbVars.toList().appendList(from);
                    to = lbArgs.toList().appendList(to);
                    t = (ClassType)substList(t, from, to);
                    for(Symbol member: s.getElements()) {
                        if (((member.flags()&PRIVATE)==0)&&(member.getKind()==ElementKind.METHOD)) {
//                            Type memberType = types.memberType(t, member);
                            MethodType methodType = member.type.asMethodType();
                            if (!isMethodExcludedFromCodeFreeRequirement(t.toString(), member.toString())) {
                                Type memberType = substList(methodType.restype, from, to);
                                memberType = substList(memberType, from, to);
                                memberType = types.upperBound(memberType);
                                // System.out.println(t+" "+member+" ----> "+memberType);
                                
//                            System.out.println("\tisObject? "+(memberType==syms.objectType)+" isCode? "+isCodeType(memberType));
                                // TODO: Do we want to allow Object? This may be good enough if we disallow casts to Code.
                                // if ((memberType==syms.objectType) || (isCodeType(memberType))) { System.out.println("! Object or Code"); result = true; return result; }
                                if (!memberType.tsym.isSubClass(syms.codeFreeType.tsym, types)) {
                                    if (isCodeType(memberType)) { 
                                        // System.out.println("    ### isCodeType");
                                        result = true;
                                        return result;
                                    }
                                    if ((memberType.tag==CLASS) && ((memberType.tsym.flags() & FINAL)==0)) {
                                        // not final class
                                        // System.out.println("    ### not final");
                                        result = true;
                                        return result;
                                    }
                                    if ((hasCodeMethods(memberType, visited) || hasCodeFields(memberType))) {
                                        // System.out.println("    ### hasCodeMethods");
                                        result = true;
                                        return result;
                                    }
                                }
                            }
//                            else {
//                                System.out.println("Excluded: "+t+" "+member);
//                            }
                        }
                    }
                    type = ((ClassSymbol)t.tsym).getSuperclass();
//                    System.out.println("superclass: "+type);
                }
                else {
                    type = Type.noType;
                }
            } while(type!=Type.noType);
//            System.out.println("done.");
            result = false;
            return result;
        }
        finally {
            // if there are no generics involved, then we can already cache the classes
            for(Type c: checked.toList()) {
                hasCodeMethodsCache.put(c.toString(), result);
//                System.out.println("checked: "+c.toString()+" "+result);
            }
        }
    }
    
    protected boolean isCodeType(Type memberType) {
        return (memberType.tsym!=null) &&
            ((memberType.tsym.name==syms.codeType.tsym.name) ||
             (memberType.tsym.name==syms.safeCodeType.tsym.name));
    }
    
    protected Type substList(Type t, List<Type> from, List<Type> to) {
//        System.out.println("\torig "+t);
//        System.out.print("\tfrom:");
//        for(Type f: from) { System.out.print(" "+f+"("+System.identityHashCode(f)+")"); }
//        System.out.print("\n\tto");
//        for(Type f: to) { System.out.print(" "+f+"("+System.identityHashCode(f)+")"); }
        for (; from.nonEmpty(); from = from.tail, to = to.tail) {
            t = types.subst(t, from, to);
        }
//        System.out.println("\nSUBST "+t);
        return t;
    }

    protected boolean isPrimitiveOrBoxedOrArrayThereof(Type type) {
        if (isPrimitiveOrBoxedOrString(type)) return true;
        if (type.tag==ARRAY) return (isPrimitiveOrBoxedOrArrayThereof(((ArrayType)type).elemtype)); 
        return false;
        // TODO: what else can we get here?
        // if (type.tag==TYPEVAR) return false;
        // throw new AssertionError("Expected either primitive, class or array for variables, was "+type.getClass().getSimpleName());
    }

    public boolean isPrimitiveOrBoxedOrString(Type type) {
        return _isPrimitiveOrBoxedOrString(type, types, syms);
    }
        
    public static boolean _isPrimitiveOrBoxedOrString(Type type, Types types, Symtab syms) {
        if (type.isPrimitive()) return true;
        if (type.tag==CLASS) {
            return ((type.tsym.equals(types.boxedClass(syms.byteType))) ||
                    (type.tsym.equals(types.boxedClass(syms.charType))) ||
                    (type.tsym.equals(types.boxedClass(syms.shortType))) ||
                    (type.tsym.equals(types.boxedClass(syms.intType))) ||
                    (type.tsym.equals(types.boxedClass(syms.longType))) ||
                    (type.tsym.equals(types.boxedClass(syms.floatType))) ||
                    (type.tsym.equals(types.boxedClass(syms.doubleType))) ||
                    (type.tsym.equals(types.boxedClass(syms.booleanType))) ||
                    (type.tsym.equals(syms.stringType.tsym)));
        }
        return false;
        // TODO: what else can we get here?
        // if (type.tag==TYPEVAR) return false;
        // throw new AssertionError("Expected either primitive, class or array for variables, was "+type.getClass().getSimpleName());
    }

    protected void cspVariableCheck(Symbol sym, JCTree tree) {
        // CSP variables must:
        // 1. be final if they are local variables or parameters
        // 2. be primitive or their corresponding boxed types, or an array thereof
        
        // check 1: local variables and parameters must be final
        if (((sym.getKind()== edu.rice.cs.mint.comp.javax.lang.model.element.ElementKind.PARAMETER) || (sym.getKind()==ElementKind.LOCAL_VARIABLE)) &&
            (sym.flags() & FINAL)==0) {
            log.error(tree.pos(), "local.var.accessed.from.brackets.needs.final", sym);
        }
        
        // check 2: must be primitive, boxed, string, or...
        if (isPrimitiveOrBoxedOrArrayThereof(sym.type)) { return; }
        
        // FIXME: revisit final classes without code fields
        // ...be final without code fields, or...
        if (sym.type.tag==CLASS) {
            ClassSymbol c = (ClassSymbol)sym.type.tsym;
            if ((c.flags() & FINAL)!=0) {
                // is final
                if ((!hasCodeFields(sym.type)) &&
                    (!hasCodeMethods(sym.type))) { return; }
            }
            // declared CodeFree
            if (c.isSubClass(syms.codeFreeType.tsym, types)) { return; }
        }
        
        log.error(tree.pos(), "csp.var.must.be.primitive.or.boxed.or.final.no.code.or.array", sym);
    }
        
    public void visitSelect(JCFieldAccess tree) {
        // Determine the expected kind of the qualifier expression.
        int skind = 0;
        if (tree.name == names._this || tree.name == names._super ||
            tree.name == names._class)
        {
            skind = TYP;
        } else {
            if ((pkind & PCK) != 0) skind = skind | PCK;
            if ((pkind & TYP) != 0) skind = skind | TYP | PCK;
            if ((pkind & (VAL | MTH)) != 0) skind = skind | VAL | TYP;
        }

        // Attribute the qualifier expression, and determine its symbol (if any).
        Type site = attribTree(tree.selected, env, skind, Infer.anyPoly);
        if ((pkind & (PCK | TYP)) == 0)
            site = capture(site); // Capture field access

        // don't allow T.class T[].class, etc
        if (skind == TYP) {
            Type elt = site;
            while (elt.tag == ARRAY)
                elt = ((ArrayType)elt).elemtype;
            if (elt.tag == TYPEVAR) {
                log.error(tree.pos(), "type.var.cant.be.deref");
                result = types.createErrorType(tree.type);
                return;
            }
        }

        // If qualifier symbol is a type or `super', assert `selectSuper'
        // for the selection. This is relevant for determining whether
        // protected symbols are accessible.
        Symbol sitesym = TreeInfo.symbol(tree.selected);
        boolean selectSuperPrev = env.info.selectSuper;
        env.info.selectSuper =
            sitesym != null &&
            sitesym.name == names._super;

        // If selected expression is polymorphic, strip
        // type parameters and remember in env.info.tvars, so that
        // they can be added later (in Attr.checkId and Infer.instantiateMethod).
        if (tree.selected.type.tag == FORALL) {
            ForAll pstype = (ForAll)tree.selected.type;
            env.info.tvars = pstype.tvars;
            site = tree.selected.type = pstype.qtype;
        }

        // Determine the symbol represented by the selection.
        env.info.varArgs = false;
        Symbol sym = selectSym(tree, site, env, pt, pkind);
        if (sym.exists() && !isType(sym) && (pkind & (PCK | TYP)) != 0) {
            site = capture(site);
            sym = selectSym(tree, site, env, pt, pkind);
        }
        boolean varArgs = env.info.varArgs;
        tree.sym = sym;

        if (site.tag == TYPEVAR && !isType(sym) && sym.kind != ERR) {
            while (site.tag == TYPEVAR) site = site.getUpperBound();
            site = capture(site);
        }

        // If that symbol is a variable, ...
        if (sym.kind == VAR) {
            VarSymbol v = (VarSymbol)sym;

            // ..., evaluate its initializer, if it has one, and check for
            // illegal forward reference.
            checkInit(tree, env, v, true);

            // If we are expecting a variable (as opposed to a value), check
            // that the variable is assignable in the current environment.
            if (pkind == VAR)
                checkAssignable(tree.pos(), v, tree.selected, env);
            
            // mgr: csp safety
            // Prevent use of this inside brackets
            if (env.enclBrackets.size()>0) {
                if (tree.sym.name==names._this) {
                    // find the environment where the bracket was created
                    ClassSymbol bracketEnclClass = null;
                    Env<AttrContext> e = env;
                    while(e!=null) {
                        if (e.tree==env.enclBrackets.head) {
                            bracketEnclClass = e.enclClass.sym;
                            break;
                        }
                        e = e.next;
                    }
                    // find out tree.sym.owner is an enclosing class of bracketEnclClass
                    Type t = bracketEnclClass.type;
                    while(t!=null) {
                        if (tree.sym.owner.type.toString().equals(t.toString())) {
                            // tree.sym.owner is an enclosing class of bracketEnclClass
                            // that means that the this belongs to a class outside the brackets
                            log.error(tree.pos(), "may.not.access.enclosing.this");
                            break;
                        }
                        t = t.getEnclosingType();
                    }
                }
            }
        }

        // Disallow selecting a type from an expression
        if (isType(sym) && (sitesym==null || (sitesym.kind&(TYP|PCK)) == 0)) {
            tree.type = check(tree.selected, pt,
                              sitesym == null ? VAL : sitesym.kind, TYP|PCK, pt);
        }

        if (isType(sitesym)) {
            if (sym.name == names._this) {
                // If `C' is the currently compiled class, check that
                // C.this' does not appear in a call to a super(...)
                if (env.info.isSelfCall &&
                    site.tsym == env.enclClass.sym) {
                    chk.earlyRefError(tree.pos(), sym);
                }
            } else {
                // Check if type-qualified fields or methods are static (JLS)
                if ((sym.flags() & STATIC) == 0 &&
                    sym.name != names._super &&
                    (sym.kind == VAR || sym.kind == MTH)) {
                    rs.access(rs.new StaticError(sym),
                              tree.pos(), site, sym.name, true);
                }
            }
        }

        // If we are selecting an instance member via a `super', ...
        if (env.info.selectSuper && (sym.flags() & STATIC) == 0) {

            // Check that super-qualified symbols are not abstract (JLS)
            rs.checkNonAbstract(tree.pos(), sym);

            if (site.isRaw()) {
                // Determine argument types for site.
                Type site1 = types.asSuper(env.enclClass.sym.type, site.tsym);
                if (site1 != null) site = site1;
            }
        }

        env.info.selectSuper = selectSuperPrev;
        result = checkId(tree, site, sym, env, pkind, pt, varArgs);
        env.info.tvars = List.nil();
    }
    //where
        /** Determine symbol referenced by a Select expression,
         *
         *  @param tree   The select tree.
         *  @param site   The type of the selected expression,
         *  @param env    The current environment.
         *  @param pt     The current prototype.
         *  @param pkind  The expected kind(s) of the Select expression.
         */
        private Symbol selectSym(JCFieldAccess tree,
                                 Type site,
                                 Env<AttrContext> env,
                                 Type pt,
                                 int pkind) {
            DiagnosticPosition pos = tree.pos();
            Name name = tree.name;

            switch (site.tag) {
            case PACKAGE:
                return rs.access(
                    rs.findIdentInPackage(env, site.tsym, name, pkind),
                    pos, site, name, true);
            case ARRAY:
            case CLASS:
                if (pt.tag == METHOD || pt.tag == FORALL) {
                    return rs.resolveQualifiedMethod(
                        pos, env, site, name, pt.getParameterTypes(), pt.getTypeArguments());
                } else if (name == names._this || name == names._super) {
                    return rs.resolveSelf(pos, env, site.tsym, name);
                } else if (name == names._class) {
                    // In this case, we have already made sure in
                    // visitSelect that qualifier expression is a type.
                    Type t = syms.classType;
                    List<Type> typeargs = allowGenerics
                        ? List.of(types.erasure(site))
                        : List.<Type>nil();
                    t = new ClassType(t.getEnclosingType(), typeargs, t.tsym);
                    return new VarSymbol(
                        STATIC | PUBLIC | FINAL, names._class, t, site.tsym);
                } else {
                    // We are seeing a plain identifier as selector.
                    Symbol sym = rs.findIdentInType(env, site, name, pkind);
                    if ((pkind & ERRONEOUS) == 0)
                        sym = rs.access(sym, pos, site, name, true);
                    return sym;
                }
            case WILDCARD:
                throw new AssertionError(tree);
            case TYPEVAR:
                // Normally, site.getUpperBound() shouldn't be null.
                // It should only happen during memberEnter/attribBase
                // when determining the super type which *must* be
                // done before attributing the type variables.  In
                // other words, we are seeing this illegal program:
                // class B<T> extends A<T.foo> {}
                Symbol sym = (site.getUpperBound() != null)
                    ? selectSym(tree, capture(site.getUpperBound()), env, pt, pkind)
                    : null;
                if (sym == null || isType(sym)) {
                    log.error(pos, "type.var.cant.be.deref");
                    return syms.errSymbol;
                } else {
                    Symbol sym2 = (sym.flags() & Flags.PRIVATE) != 0 ?
                        rs.new AccessError(env, site, sym) :
                                sym;
                    rs.access(sym2, pos, site, name, true);
                    return sym;
                }
            case ERROR:
                // preserve identifier names through errors
                return types.createErrorType(name, site.tsym, site).tsym;
            default:
                // The qualifier expression is of a primitive type -- only
                // .class is allowed for these.
                if (name == names._class) {
                    // In this case, we have already made sure in Select that
                    // qualifier expression is a type.
                    Type t = syms.classType;
                    Type arg = types.boxedClass(site).type;
                    t = new ClassType(t.getEnclosingType(), List.of(arg), t.tsym);
                    return new VarSymbol(
                        STATIC | PUBLIC | FINAL, names._class, t, site.tsym);
                } else {
                    log.error(pos, "cant.deref", site);
                    return syms.errSymbol;
                }
            }
        }

        /** Determine type of identifier or select expression and check that
         *  (1) the referenced symbol is not deprecated
         *  (2) the symbol's type is safe (@see checkSafe)
         *  (3) if symbol is a variable, check that its type and kind are
         *      compatible with the prototype and protokind.
         *  (4) if symbol is an instance field of a raw type,
         *      which is being assigned to, issue an unchecked warning if its
         *      type changes under erasure.
         *  (5) if symbol is an instance method of a raw type, issue an
         *      unchecked warning if its argument types change under erasure.
         *  If checks succeed:
         *    If symbol is a constant, return its constant type
         *    else if symbol is a method, return its result type
         *    otherwise return its type.
         *  Otherwise return errType.
         *
         *  @param tree       The syntax tree representing the identifier
         *  @param site       If this is a select, the type of the selected
         *                    expression, otherwise the type of the current class.
         *  @param sym        The symbol representing the identifier.
         *  @param env        The current environment.
         *  @param pkind      The set of expected kinds.
         *  @param pt         The expected type.
         */
        Type checkId(JCTree tree,
                     Type site,
                     Symbol sym,
                     Env<AttrContext> env,
                     int pkind,
                     Type pt,
                     boolean useVarargs) {
            if (pt.isErroneous()) return types.createErrorType(site);
            Type owntype; // The computed type of this identifier occurrence.
            switch (sym.kind) {
            case TYP:
                // For types, the computed type equals the symbol's type,
                // except for two situations:
                owntype = sym.type;
                if (owntype.tag == CLASS) {
                    Type ownOuter = owntype.getEnclosingType();

                    // (a) If the symbol's type is parameterized, erase it
                    // because no type parameters were given.
                    // We recover generic outer type later in visitTypeApply.
                    if (owntype.tsym.type.getTypeArguments().nonEmpty()) {
                        owntype = types.erasure(owntype);
                    }

                    // (b) If the symbol's type is an inner class, then
                    // we have to interpret its outer type as a superclass
                    // of the site type. Example:
                    //
                    // class Tree<A> { class Visitor { ... } }
                    // class PointTree extends Tree<Point> { ... }
                    // ...PointTree.Visitor...
                    //
                    // Then the type of the last expression above is
                    // Tree<Point>.Visitor.
                    else if (ownOuter.tag == CLASS && site != ownOuter) {
                        Type normOuter = site;
                        if (normOuter.tag == CLASS)
                            normOuter = types.asEnclosingSuper(site, ownOuter.tsym);
                        if (normOuter == null) // perhaps from an import
                            normOuter = types.erasure(ownOuter);
                        if (normOuter != ownOuter)
                            owntype = new ClassType(
                                normOuter, List.<Type>nil(), owntype.tsym);
                    }
                }
                break;
            case VAR:
                VarSymbol v = (VarSymbol)sym;
                // Test (4): if symbol is an instance field of a raw type,
                // which is being assigned to, issue an unchecked warning if
                // its type changes under erasure.
                if (allowGenerics &&
                    pkind == VAR &&
                    v.owner.kind == TYP &&
                    (v.flags() & STATIC) == 0 &&
                    (site.tag == CLASS || site.tag == TYPEVAR)) {
                    Type s = types.asOuterSuper(site, v.owner);
                    if (s != null &&
                        s.isRaw() &&
                        !types.isSameType(v.type, v.erasure(types))) {
                        chk.warnUnchecked(tree.pos(),
                                          "unchecked.assign.to.var",
                                          v, s);
                    }
                }
                // The computed type of a variable is the type of the
                // variable symbol, taken as a member of the site type.
                owntype = (sym.owner.kind == TYP &&
                           sym.name != names._this && sym.name != names._super)
                    ? types.memberType(site, sym)
                    : sym.type;

                if (env.info.tvars.nonEmpty()) {
                    Type owntype1 = new ForAll(env.info.tvars, owntype);
                    for (List<Type> l = env.info.tvars; l.nonEmpty(); l = l.tail)
                        if (!owntype.contains(l.head)) {
                            log.error(tree.pos(), "undetermined.type", owntype1);
                            owntype1 = types.createErrorType(owntype1);
                        }
                    owntype = owntype1;
                }

                // If the variable is a constant, record constant value in
                // computed type.
                if (v.getConstValue() != null && isStaticReference(tree))
                    owntype = owntype.constType(v.getConstValue());

                if (pkind == VAL) {
                    owntype = capture(owntype); // capture "names as expressions"
                }
                break;
            case MTH: {
                JCMethodInvocation app = (JCMethodInvocation)env.tree;
                owntype = checkMethod(site, sym, env, app.args,
                                      pt.getParameterTypes(), pt.getTypeArguments(),
                                      env.info.varArgs);
                break;
            }
            case PCK: case ERR:
                owntype = sym.type;
                break;
            default:
                throw new AssertionError("unexpected kind: " + sym.kind +
                                         " in tree " + tree);
            }

            // Test (1): emit a `deprecation' warning if symbol is deprecated.
            // (for constructors, the error was given when the constructor was
            // resolved)
            if (sym.name != names.init &&
                (sym.flags() & DEPRECATED) != 0 &&
                (env.info.scope.owner.flags() & DEPRECATED) == 0 &&
                sym.outermostClass() != env.info.scope.owner.outermostClass())
                chk.warnDeprecated(tree.pos(), sym);

            if ((sym.flags() & PROPRIETARY) != 0)
                log.strictWarning(tree.pos(), "sun.proprietary", sym);

            // Test (3): if symbol is a variable, check that its type and
            // kind are compatible with the prototype and protokind.
            return check(tree, owntype, sym.kind, pkind, pt);
        }

        /** Check that variable is initialized and evaluate the variable's
         *  initializer, if not yet done. Also check that variable is not
         *  referenced before it is defined.
         *  @param tree    The tree making up the variable reference.
         *  @param env     The current environment.
         *  @param v       The variable's symbol.
         */
        private void checkInit(JCTree tree,
                               Env<AttrContext> env,
                               VarSymbol v,
                               boolean onlyWarning) {
//          System.err.println(v + " " + ((v.flags() & STATIC) != 0) + " " +
//                             tree.pos + " " + v.pos + " " +
//                             Resolve.isStatic(env));//DEBUG

            // A forward reference is diagnosed if the declaration position
            // of the variable is greater than the current tree position
            // and the tree and variable definition occur in the same class
            // definition.  Note that writes don't count as references.
            // This check applies only to class and instance
            // variables.  Local variables follow different scope rules,
            // and are subject to definite assignment checking.
            if ((env.info.enclVar == v || v.pos > tree.pos) &&
                v.owner.kind == TYP &&
                canOwnInitializer(env.info.scope.owner) &&
                v.owner == env.info.scope.owner.enclClass() &&
                ((v.flags() & STATIC) != 0) == Resolve.isStatic(env) &&
                (env.tree.getTag() != JCTree.ASSIGN ||
                 TreeInfo.skipParens(((JCAssign) env.tree).lhs) != tree)) {
                String suffix = (env.info.enclVar == v) ?
                                "self.ref" : "forward.ref";
                if (!onlyWarning || isStaticEnumField(v)) {
                    log.error(tree.pos(), "illegal." + suffix);
                } else if (useBeforeDeclarationWarning) {
                    log.warning(tree.pos(), suffix, v);
                }
            }

            v.getConstValue(); // ensure initializer is evaluated

            checkEnumInitializer(tree, env, v);
        }

        /**
         * Check for illegal references to static members of enum.  In
         * an enum type, constructors and initializers may not
         * reference its static members unless they are constant.
         *
         * @param tree    The tree making up the variable reference.
         * @param env     The current environment.
         * @param v       The variable's symbol.
         * @see JLS 3rd Ed. (8.9 Enums)
         */
        private void checkEnumInitializer(JCTree tree, Env<AttrContext> env, VarSymbol v) {
            // JLS 3rd Ed.:
            //
            // "It is a compile-time error to reference a static field
            // of an enum type that is not a compile-time constant
            // (15.28) from constructors, instance initializer blocks,
            // or instance variable initializer expressions of that
            // type. It is a compile-time error for the constructors,
            // instance initializer blocks, or instance variable
            // initializer expressions of an enum constant e to refer
            // to itself or to an enum constant of the same type that
            // is declared to the right of e."
            if (isStaticEnumField(v)) {
                ClassSymbol enclClass = env.info.scope.owner.enclClass();

                if (enclClass == null || enclClass.owner == null)
                    return;

                // See if the enclosing class is the enum (or a
                // subclass thereof) declaring v.  If not, this
                // reference is OK.
                if (v.owner != enclClass && !types.isSubtype(enclClass.type, v.owner.type))
                    return;

                // If the reference isn't from an initializer, then
                // the reference is OK.
                if (!Resolve.isInitializer(env))
                    return;

                log.error(tree.pos(), "illegal.enum.static.ref");
            }
        }

        /** Is the given symbol a static, non-constant field of an Enum?
         *  Note: enum literals should not be regarded as such
         */
        private boolean isStaticEnumField(VarSymbol v) {
            return Flags.isEnum(v.owner) &&
                   Flags.isStatic(v) &&
                   !Flags.isConstant(v) &&
                   v.name != names._class;
        }

        /** Can the given symbol be the owner of code which forms part
         *  if class initialization? This is the case if the symbol is
         *  a type or field, or if the symbol is the synthetic method.
         *  owning a block.
         */
        private boolean canOwnInitializer(Symbol sym) {
            return
                (sym.kind & (VAR | TYP)) != 0 ||
                (sym.kind == MTH && (sym.flags() & BLOCK) != 0);
        }

    Warner noteWarner = new Warner();

    /**
     * Check that method arguments conform to its instantation.
     **/
    public Type checkMethod(Type site,
                            Symbol sym,
                            Env<AttrContext> env,
                            final List<JCExpression> argtrees,
                            List<Type> argtypes,
                            List<Type> typeargtypes,
                            boolean useVarargs) {
        // Test (5): if symbol is an instance method of a raw type, issue
        // an unchecked warning if its argument types change under erasure.
        if (allowGenerics &&
            (sym.flags() & STATIC) == 0 &&
            (site.tag == CLASS || site.tag == TYPEVAR)) {
            Type s = types.asOuterSuper(site, sym.owner);
            if (s != null && s.isRaw() &&
                !types.isSameTypes(sym.type.getParameterTypes(),
                                   sym.erasure(types).getParameterTypes())) {
                chk.warnUnchecked(env.tree.pos(),
                                  "unchecked.call.mbr.of.raw.type",
                                  sym, s);
            }
        }

        // Compute the identifier's instantiated type.
        // For methods, we need to compute the instance type by
        // Resolve.instantiate from the symbol's type as well as
        // any type arguments and value arguments.
        noteWarner.warned = false;
        Type owntype = rs.instantiate(env,
                                      site,
                                      sym,
                                      argtypes,
                                      typeargtypes,
                                      true,
                                      useVarargs,
                                      noteWarner);
        boolean warned = noteWarner.warned;

        // If this fails, something went wrong; we should not have
        // found the identifier in the first place.
        if (owntype == null) {
            if (!pt.isErroneous())
                log.error(env.tree.pos(),
                          "internal.error.cant.instantiate",
                          sym, site,
                          Type.toString(pt.getParameterTypes()));
            owntype = types.createErrorType(site);
        } else {
            // System.out.println("call   : " + env.tree);
            // System.out.println("method : " + owntype);
            // System.out.println("actuals: " + argtypes);
            List<Type> formals = owntype.getParameterTypes();
            Type last = useVarargs ? formals.last() : null;
            if (sym.name==names.init &&
                sym.owner == syms.enumSym)
                formals = formals.tail.tail;
            List<JCExpression> args = argtrees;
            while (formals.head != last) {
                JCTree arg = args.head;
                Warner warn = chk.convertWarner(arg.pos(), arg.type, formals.head);
                assertConvertible(arg, arg.type, formals.head, warn);
                warned |= warn.warned;
                args = args.tail;
                formals = formals.tail;
            }
            if (useVarargs) {
                Type varArg = types.elemtype(last);
                while (args.tail != null) {
                    JCTree arg = args.head;
                    Warner warn = chk.convertWarner(arg.pos(), arg.type, varArg);
                    assertConvertible(arg, arg.type, varArg, warn);
                    warned |= warn.warned;
                    args = args.tail;
                }
            } else if ((sym.flags() & VARARGS) != 0 && allowVarargs) {
                // non-varargs call to varargs method
                Type varParam = owntype.getParameterTypes().last();
                Type lastArg = argtypes.last();
                if (types.isSubtypeUnchecked(lastArg, types.elemtype(varParam)) &&
                    !types.isSameType(types.erasure(varParam), types.erasure(lastArg)))
                    log.warning(argtrees.last().pos(), "inexact.non-varargs.call",
                                types.elemtype(varParam),
                                varParam);
            }

            if (warned && sym.type.tag == FORALL) {
                chk.warnUnchecked(env.tree.pos(),
                                  "unchecked.meth.invocation.applied",
                                  kindName(sym),
                                  sym.name,
                                  rs.methodArguments(sym.type.getParameterTypes()),
                                  rs.methodArguments(argtypes),
                                  kindName(sym.location()),
                                  sym.location());
                owntype = new MethodType(owntype.getParameterTypes(),
                                         types.erasure(owntype.getReturnType()),
                                         owntype.getThrownTypes(),
                                         syms.methodClass);
            }
            if (useVarargs) {
                JCTree tree = env.tree;
                Type argtype = owntype.getParameterTypes().last();
                if (!types.isReifiable(argtype))
                    chk.warnUnchecked(env.tree.pos(),
                                      "unchecked.generic.array.creation",
                                      argtype);
                Type elemtype = types.elemtype(argtype);
                switch (tree.getTag()) {
                case JCTree.APPLY:
                    ((JCMethodInvocation) tree).varargsElement = elemtype;
                    break;
                case JCTree.NEWCLASS:
                    ((JCNewClass) tree).varargsElement = elemtype;
                    break;
                default:
                    throw new AssertionError(""+tree);
                }
            }
        }
        return owntype;
    }

    private void assertConvertible(JCTree tree, Type actual, Type formal, Warner warn) {
        if (types.isConvertible(actual, formal, warn))
            return;

        if (formal.isCompound()
            && types.isSubtype(actual, types.supertype(formal))
            && types.isSubtypeUnchecked(actual, types.interfaces(formal), warn))
            return;

        if (false) {
            // TODO: make assertConvertible work
            chk.typeError(tree.pos(), diags.fragment("incompatible.types"), actual, formal);
            throw new AssertionError("Tree: " + tree
                                     + " actual:" + actual
                                     + " formal: " + formal);
        }
    }

    public void visitLiteral(JCLiteral tree) {
        result = check(
            tree, litType(tree.typetag).constType(tree.value), VAL, pkind, pt);
    }
    //where
    /** Return the type of a literal with given type tag.
     */
    Type litType(int tag) {
        return (tag == TypeTags.CLASS) ? syms.stringType : syms.typeOfTag[tag];
    }

    public void visitTypeIdent(JCPrimitiveTypeTree tree) {
        result = check(tree, syms.typeOfTag[tree.typetag], TYP, pkind, pt);
    }

    public void visitTypeArray(JCArrayTypeTree tree) {
        Type etype = attribType(tree.elemtype, env);
        Type type = new ArrayType(etype, syms.arrayClass);
        result = check(tree, type, TYP, pkind, pt);
    }

    /** Visitor method for parameterized types.
     *  Bound checking is left until later, since types are attributed
     *  before supertype structure is completely known
     */
    public void visitTypeApply(JCTypeApply tree) {
        Type owntype = types.createErrorType(tree.type);

        // Attribute functor part of application and make sure it's a class.
        Type clazztype = chk.checkClassType(tree.clazz.pos(), attribType(tree.clazz, env));

        // Attribute type parameters
        List<Type> actuals = attribTypes(tree.arguments, env);

        if (clazztype.tag == CLASS) {
            List<Type> formals = clazztype.tsym.type.getTypeArguments();

            if (actuals.length() == formals.length()) {
                List<Type> a = actuals;
                List<Type> f = formals;
                while (a.nonEmpty()) {
                    a.head = a.head.withTypeVar(f.head);
                    a = a.tail;
                    f = f.tail;
                }
                // Compute the proper generic outer
                Type clazzOuter = clazztype.getEnclosingType();
                if (clazzOuter.tag == CLASS) {
                    Type site;
                    if (tree.clazz.getTag() == JCTree.IDENT) {
                        site = env.enclClass.sym.type;
                    } else if (tree.clazz.getTag() == JCTree.SELECT) {
                        site = ((JCFieldAccess) tree.clazz).selected.type;
                    } else throw new AssertionError(""+tree);
                    if (clazzOuter.tag == CLASS && site != clazzOuter) {
                        if (site.tag == CLASS)
                            site = types.asOuterSuper(site, clazzOuter.tsym);
                        if (site == null)
                            site = types.erasure(clazzOuter);
                        clazzOuter = site;
                    }
                }
                owntype = new ClassType(clazzOuter, actuals, clazztype.tsym);
            } else {
                if (formals.length() != 0) {
                    log.error(tree.pos(), "wrong.number.type.args",
                              Integer.toString(formals.length()));
                } else {
                    log.error(tree.pos(), "type.doesnt.take.params", clazztype.tsym);
                }
                owntype = types.createErrorType(tree.type);
            }
            
            // mgr: prevent instantiation of type parameters to Code, except when in Code
            // List<Code> or List<Code<Integer>> illegal
            // Code<Code<Integer>> legal
            // Code<List<Code>> illegal
            // NOTE: We don't do that anymore
            // if (!types.isSubtypeUnchecked(clazztype,syms.codeType)) {
            //     // A<T> is not Code<T> -- T may not be Code 
            //     boolean hasCodeArg = false;
            //     for(JCExpression arg: tree.arguments) {
            //         if (arg instanceof JCIdent) {
            //             JCIdent t = (JCIdent)arg;
            //             Type ttype = null;
            //             if (t.sym instanceof ClassSymbol) {
            //                 ttype = ((ClassSymbol)t.sym).type;
            //             }
            //             else if (t.sym instanceof TypeSymbol) {
            //                 TypeSymbol tsym = (TypeSymbol)t.sym;
            //                 ttype = tsym.type;
            //             }
            //             else throw new AssertionError("Expected ClassSymbol or TypeSymbol as type argument");
            //             if (types.isSubtypeUnchecked(ttype,syms.codeType)) {
            //                 hasCodeArg = true;
            //                 break;
            //             }
            //         }
            //         else if (arg instanceof JCTypeApply) {
            //             JCTypeApply tapp = (JCTypeApply)arg;
            //             JCIdent t = (JCIdent)tapp.clazz;
            //             Type ttype = ((ClassSymbol)t.sym).type;
            //             if (types.isSubtypeUnchecked(ttype,syms.codeType)) {
            //                 hasCodeArg = true;
            //                 break;
            //             }
            //         }
            //     }
            //     if (hasCodeArg) {
            //         log.error(tree.pos(), "may.not.instantiate.to.code");
            //     }
            // }
        }
        result = check(tree, owntype, TYP, pkind, pt);        
    }

    public void visitTypeParameter(JCTypeParameter tree) {
        TypeVar a = (TypeVar)tree.type;
        Set<Type> boundSet = new HashSet<Type>();
        if (a.bound.isErroneous())
            return;
        List<Type> bs = types.getBounds(a);
        if (tree.bounds.nonEmpty()) {
            // accept class or interface or typevar as first bound.
            Type b = checkBase(bs.head, tree.bounds.head, env, false, false, false);
            boundSet.add(types.erasure(b));
            if (b.isErroneous()) {
                a.bound = b;
            }
            else if (b.tag == TYPEVAR) {
                // if first bound was a typevar, do not accept further bounds.
                if (tree.bounds.tail.nonEmpty()) {
                    log.error(tree.bounds.tail.head.pos(),
                              "type.var.may.not.be.followed.by.other.bounds");
                    tree.bounds = List.of(tree.bounds.head);
                    a.bound = bs.head;
                }
            } else {
                // if first bound was a class or interface, accept only interfaces
                // as further bounds.
                for (JCExpression bound : tree.bounds.tail) {
                    bs = bs.tail;
                    Type i = checkBase(bs.head, bound, env, false, true, false);
                    if (i.isErroneous())
                        a.bound = i;
                    else if (i.tag == CLASS)
                        chk.checkNotRepeated(bound.pos(), types.erasure(i), boundSet);
                }
            }
        }
        bs = types.getBounds(a);

        // in case of multiple bounds ...
        if (bs.length() > 1) {
            // ... the variable's bound is a class type flagged COMPOUND
            // (see comment for TypeVar.bound).
            // In this case, generate a class tree that represents the
            // bound class, ...
            JCTree extending;
            List<JCExpression> implementing;
            if ((bs.head.tsym.flags() & INTERFACE) == 0) {
                extending = tree.bounds.head;
                implementing = tree.bounds.tail;
            } else {
                extending = null;
                implementing = tree.bounds;
            }
            JCClassDecl cd = make.at(tree.pos).ClassDef(
                make.Modifiers(PUBLIC | ABSTRACT),
                tree.name, List.<JCTypeParameter>nil(),
                extending, implementing, List.<JCTree>nil());

            ClassSymbol c = (ClassSymbol)a.getUpperBound().tsym;
            assert (c.flags() & COMPOUND) != 0;
            cd.sym = c;
            c.sourcefile = env.toplevel.sourcefile;

            // ... and attribute the bound class
            c.flags_field |= UNATTRIBUTED;
            Env<AttrContext> cenv = enter.classEnv(cd, env);
            enter.typeEnvs.put(c, cenv);
        }
    }


    public void visitWildcard(JCWildcard tree) {
        //- System.err.println("visitWildcard("+tree+");");//DEBUG
        Type type = (tree.kind.kind == BoundKind.UNBOUND)
            ? syms.objectType
            : attribType(tree.inner, env);
        result = check(tree, new WildcardType(chk.checkRefType(tree.pos(), type),
                                              tree.kind.kind,
                                              syms.boundClass),
                       TYP, pkind, pt);
    }

    public void visitAnnotation(JCAnnotation tree) {
        log.error(tree.pos(), "annotation.not.valid.for.type", pt);
        result = tree.type = syms.errType;
    }

    public void visitErroneous(JCErroneous tree) {
        if (tree.errs != null)
            for (JCTree err : tree.errs)
                attribTree(err, env, ERR, pt);
        result = tree.type = syms.errType;
    }

    /** Default visitor method for all other trees.
     */
    public void visitTree(JCTree tree) {
        throw new AssertionError();
    }

    /** Main method: attribute class definition associated with given class symbol.
     *  reporting completion failures at the given position.
     *  @param pos The source position at which completion errors are to be
     *             reported.
     *  @param c   The class symbol whose definition will be attributed.
     */
    public void attribClass(DiagnosticPosition pos, ClassSymbol c) {
        try {
            annotate.flush();
            attribClass(c);
        } catch (CompletionFailure ex) {
            chk.completionError(pos, ex);
        }
    }

    /** Attribute class definition associated with given class symbol.
     *  @param c   The class symbol whose definition will be attributed.
     */
    void attribClass(ClassSymbol c) throws CompletionFailure {
        if (c.type.tag == ERROR) return;

        // Check for cycles in the inheritance graph, which can arise from
        // ill-formed class files.
        chk.checkNonCyclic(null, c.type);

        Type st = types.supertype(c.type);
        if ((c.flags_field & Flags.COMPOUND) == 0) {
            // First, attribute superclass.
            if (st.tag == CLASS)
                attribClass((ClassSymbol)st.tsym);

            // Next attribute owner, if it is a class.
            if (c.owner.kind == TYP && c.owner.type.tag == CLASS)
                attribClass((ClassSymbol)c.owner);
        }

        // The previous operations might have attributed the current class
        // if there was a cycle. So we test first whether the class is still
        // UNATTRIBUTED.
        if ((c.flags_field & UNATTRIBUTED) != 0) {
            c.flags_field &= ~UNATTRIBUTED;

            // Get environment current at the point of class definition.
            Env<AttrContext> env = enter.typeEnvs.get(c);

            // The info.lint field in the envs stored in enter.typeEnvs is deliberately uninitialized,
            // because the annotations were not available at the time the env was created. Therefore,
            // we look up the environment chain for the first enclosing environment for which the
            // lint value is set. Typically, this is the parent env, but might be further if there
            // are any envs created as a result of TypeParameter nodes.
            Env<AttrContext> lintEnv = env;
            while (lintEnv.info.lint == null)
                lintEnv = lintEnv.next;

            // Having found the enclosing lint value, we can initialize the lint value for this class
            env.info.lint = lintEnv.info.lint.augment(c.attributes_field, c.flags());

            Lint prevLint = chk.setLint(env.info.lint);
            edu.rice.cs.mint.comp.javax.tools.JavaFileObject prev = log.useSource(c.sourcefile);

            try {
                // java.lang.Enum may not be subclassed by a non-enum
                if (st.tsym == syms.enumSym &&
                    ((c.flags_field & (Flags.ENUM|Flags.COMPOUND)) == 0))
                    log.error(env.tree.pos(), "enum.no.subclassing");

                // Enums may not be extended by source-level classes
                if (st.tsym != null &&
                    ((st.tsym.flags_field & Flags.ENUM) != 0) &&
                    ((c.flags_field & (Flags.ENUM | Flags.COMPOUND)) == 0) &&
                    !target.compilerBootstrap(c)) {
                    log.error(env.tree.pos(), "enum.types.not.extensible");
                }
                attribClassBody(env, c);

                chk.checkDeprecatedAnnotation(env.tree.pos(), c);
            } finally {
                log.useSource(prev);
                chk.setLint(prevLint);
            }

            
            // if this class (or a superclass or interface) implements CodeFree,
            // check to make sure there are no fields that can contain code
            // TODO
            if (c.isSubClass(syms.codeFreeType.tsym, types)) {
                // class declared CodeFree
                // System.out.println("Class declared code-free: "+c);
                if (!(env.tree instanceof JCClassDecl)) {
                    throw new AssertionError("Expected a JCClassDecl here, was "+env.tree.getClass());
                }
                JCClassDecl tree = (JCClassDecl)env.tree;
                for (List<JCTree> l = tree.defs; l.nonEmpty(); l = l.tail) {
                    JCTree def = l.head;
                    // System.out.println("\t"+def.getClass().getName()+": "+def);
                    if (def instanceof JCVariableDecl) {
                        JCVariableDecl vd = (JCVariableDecl)def;
                        // System.out.println("\t\t"+vd.getType().getClass()+" "+vd.getType());
                        if (canContainCode(vd.getType(), new HashSet<String>())) {
                            log.error(def.pos(), "codefree.class.had.non.codefree.member",
                                      c.toString(), vd.name.toString());
                        }
                    }
                    else if (def instanceof JCMethodDecl) {
                        JCMethodDecl md = (JCMethodDecl)def;
                        if (md.getReturnType()!=null) {
                            // System.out.println("\t\t"+md.getReturnType().getClass()+" "+md.getReturnType());
                            if (canContainCode(md.getReturnType(), new HashSet<String>())) {
                                log.error(def.pos(), "codefree.class.had.non.codefree.member",
                                          c.toString(), md.name.toString());
                            }
                        }
                    }
                }
            }
        }
    }

//    boolean isClassExcludedFromCodeFreeRequirement(String type) {
//        if (type.equals("sun.reflect.ConstructorAccessor") ||
//            type.equals("sun.reflect.FieldAccessor") ||
//            type.equals("sun.reflect.MethodAccessor") ||
//            type.equals("java.lang.reflect.Field") ||
//            type.equals("java.lang.reflect.Method") ||
//            type.equals("java.lang.reflect.Member") ||
//            type.equals("sun.reflect.ConstantPool") ||
//            type.equals("java.lang.Package") ||
//            type.equals("java.lang.ClassLoader") ||
//            type.equals("java.security.ProtectionDomain") ||
//            false) {
//            // TODO: What about these classes? Can they cause scope extrusion in practice?
//            return true;
//        }
//        return false;
//    }
    
    boolean isMethodExcludedFromCodeFreeRequirement(String type, String method) {
        if ((method.equals("clone()")) ||
            (method.equals("getClass()")) ||
            (type.equals("java.lang.Class") && method.equals("getResource(java.lang.String)")) ||
            false) {
            // TODO: What about these methods? Can they cause scope extrusion in practice?
            return true;
        }
        return false;
    }
    
    boolean canContainCode(JCTree type, HashSet<String> visited) {
        if (type instanceof JCPrimitiveTypeTree) {
            // can't possibly contain code
            return false;
        }
        else if (type instanceof JCIdent) {
            JCIdent ci = (JCIdent)type;
            if (canContainCode(ci.sym, visited)) return true;
        }
        else if (type instanceof JCTypeApply) {
            JCTypeApply ta = (JCTypeApply)type;
            if (canContainCode(ta.clazz, visited)) return true;
            for (JCExpression arg: ta.arguments) {
                if (canContainCode(arg, visited)) return true;
            }
        }
        else if (type instanceof JCArrayTypeTree) {
            JCArrayTypeTree at = (JCArrayTypeTree)type;
            if (canContainCode(at.elemtype, visited)) return true;
        }
        else {
            throw new AssertionError("Expected JCIdent, JCTypeApply or JCArrayTypeTree, was "+type.getClass().getName());
        }
        return false;
    }

    boolean canContainCode(Symbol sym, HashSet<String> visited) {
        if (!(sym instanceof ClassSymbol)) return false;
        ClassSymbol csym = (ClassSymbol)sym;
        
        if (csym.isSubClass(syms.codeFreeType.tsym, types)) return false;
        
        // must be primitive, boxed, string, or...
        if (isPrimitiveOrBoxedOrArrayThereof(sym.type)) { return false; }
        
        // ...be final without code fields
        if ((csym.flags() & FINAL)!=0) {
            // is final
            if ((!hasCodeFields(csym.type)) &&
                (!hasCodeMethods(csym.type))) { return false; }
        }
        return true;
//        if (visited.contains(csym.type.toString())) return false;
//
//        // System.out.println("visited "+csym.type.toString());
//        visited.add(csym.type.toString()); // now this type has been visited
//        
//        if (isClassExcludedFromCodeFreeRequirement(csym.type.toString())) return false;
//        
//        if (csym.equals(syms.objectType.tsym)) {
//            // type is Object, can contain code!
//            // System.out.println("########\t"+csym);
//            return true;
//        }
//        else if (csym.isSubClass(syms.codeType.tsym, types)) {
//            // type is subclass of Code, can contain code!
//            // System.out.println("########\t"+csym);
//            return true;
//        }
//        else {
//            // it's not Object or Code
//            // but we still have to make sure that this class doesn't itself contain Code as a field
//            // i.e. it either has to implement CodeFree
//            // or it has to satisfy the old definition of code-free (final)
//            // System.out.println("\t\tscope: "+csym.members());
//            if ((csym.members()!=null) && (canMembersContainCode(env, csym.type, csym, visited))) {
//                // System.out.println("########\t"+csym);
//                return true;
//            }
//        }
//        return false;
    }
    
//    boolean canMembersContainCode(Env<AttrContext> env,
//                                  Type site,
//                                  TypeSymbol c,
//                                  HashSet<String> visited) {
//        while (c.type.tag == TYPEVAR)
//            c = c.type.getUpperBound().tsym;
//        Scope.Entry e = c.members().elems;
//        while (e != null) {
//            // System.out.println("\t\tmember "+c+" "+e.sym);
//            if (e.sym.kind == VAR && (e.sym.flags_field & SYNTHETIC) == 0) {
//                if (canContainCode(e.sym.type.tsym, visited)) return true;
//            }
//            else if (e.sym.kind == MTH && (e.sym.flags_field & SYNTHETIC) == 0) {
//                if (e.sym.type instanceof MethodType) {
//                    MethodType mt = (MethodType)e.sym.type;
//                    if (!isMethodExcludedFromCodeFreeRequirement(c.toString(), e.sym.toString())) {
//                        if (canContainCode(mt.getReturnType().tsym, visited)) {
//                            System.out.println(c+" "+e.sym);
//                            return true;
//                        }
//                    }
//                }
//                else if (e.sym.type instanceof ForAll) {
//                    ForAll fa = (ForAll)e.sym.type;
//                    MethodType mt = fa.asMethodType();
//                    if (canContainCode(mt.getReturnType().tsym, visited)) return true;
//                }
//                else {
//                    throw new AssertionError("Expected MethodType or ForAll, was "+e.sym.type.getClass().getName());
//                }
//            }
//
//            e = e.sibling;
//        }
//        Type st = types.supertype(c.type);
//        if (st != null && (st.tag == CLASS || st.tag == TYPEVAR)) {
//            if (canMembersContainCode(env, site, st.tsym, visited)) return true;
//        }
//        for (Type it: types.interfaces(c.type)) {
//            if (canMembersContainCode(env, site, it.tsym, visited)) return true;
//        }
//        return false;
//    }


    public void visitImport(JCImport tree) {
        // nothing to do
    }
    
    // mgr: added let expressions
    public void visitLetExpr(LetExpr tree) {
        attribStats(tree.defs, env);
        Type t = attribExpr(tree.expr, env);
        result = check(tree, t, pkind, pkind, pt);
    }

    /** Finish the attribution of a class. */
    private void attribClassBody(Env<AttrContext> env, ClassSymbol c) {
        JCClassDecl tree = (JCClassDecl)env.tree;
        assert c == tree.sym;

        // Validate annotations
        chk.validateAnnotations(tree.mods.annotations, c);

        // Validate type parameters, supertype and interfaces.
        attribBounds(tree.typarams);
        chk.validate(tree.typarams, env);
        chk.validate(tree.extending, env);
        chk.validate(tree.implementing, env);

        // If this is a non-abstract class, check that it has no abstract
        // methods or unimplemented methods of an implemented interface.
        if ((c.flags() & (ABSTRACT | INTERFACE)) == 0) {
            if (!relax)
                chk.checkAllDefined(tree.pos(), c);
        }

        if ((c.flags() & ANNOTATION) != 0) {
            if (tree.implementing.nonEmpty())
                log.error(tree.implementing.head.pos(),
                          "cant.extend.intf.annotation");
            if (tree.typarams.nonEmpty())
                log.error(tree.typarams.head.pos(),
                          "intf.annotation.cant.have.type.params");
        } else {
            // Check that all extended classes and interfaces
            // are compatible (i.e. no two define methods with same arguments
            // yet different return types).  (JLS 8.4.6.3)
            chk.checkCompatibleSupertypes(tree.pos(), c.type);
        }

        // Check that class does not import the same parameterized interface
        // with two different argument lists.
        chk.checkClassBounds(tree.pos(), c.type);

        tree.type = c.type;

        boolean assertsEnabled = false;
        assert assertsEnabled = true;
        if (assertsEnabled) {
            for (List<JCTypeParameter> l = tree.typarams;
                 l.nonEmpty(); l = l.tail)
                assert env.info.scope.lookup(l.head.name).scope != null;
        }

        // Check that a generic class doesn't extend Throwable
        if (!c.type.allparams().isEmpty() && types.isSubtype(c.type, syms.throwableType))
            log.error(tree.extending.pos(), "generic.throwable");

        // Check that all methods which implement some
        // method conform to the method they implement.
        chk.checkImplementations(tree);

        for (List<JCTree> l = tree.defs; l.nonEmpty(); l = l.tail) {
            // Attribute declaration
            attribStat(l.head, env);
            // Check that declarations in inner classes are not static (JLS 8.1.2)
            // Make an exception for static constants.
            if (c.owner.kind != PCK &&
                ((c.flags() & STATIC) == 0 || c.name == names.empty) &&
                (TreeInfo.flags(l.head) & (STATIC | INTERFACE)) != 0) {
                Symbol sym = null;
                if (l.head.getTag() == JCTree.VARDEF) sym = ((JCVariableDecl) l.head).sym;
                if (sym == null ||
                    sym.kind != VAR ||
                    ((VarSymbol) sym).getConstValue() == null)
                    log.error(l.head.pos(), "icls.cant.have.static.decl");
            }
        }

        // Check for cycles among non-initial constructors.
        chk.checkCyclicConstructors(tree);

        // Check for cycles among annotation elements.
        chk.checkNonCyclicElements(tree);

        // Check for proper use of serialVersionUID
        if (env.info.lint.isEnabled(Lint.LintCategory.SERIAL) &&
            isSerializable(c) &&
            (c.flags() & Flags.ENUM) == 0 &&
            (c.flags() & ABSTRACT) == 0) {
            checkSerialVersionUID(tree, c);
        }
            
        // mgr: escape safety
        // If the class has at least one separable ctor, then calling
        // non-separable methods or ctors from (static) initializers is illegal
        if (tree.containsSeparableCtors && tree.containsNonSeparableInitializers) {
            // error: calling non-separable methods or ctors even though class contains at least one separable ctor
            for(JCTree d: tree.defs) {
                if (d instanceof JCMethodDecl) {
                    JCMethodDecl m = (JCMethodDecl)d;
                    if (m.name == names.init) {
                        boolean isCtorSeparable = ((m.sym.flags() & SEPARABLE)!=0);
                        if (isCtorSeparable) {
                            log.error(m.pos(), "separable.ctor.but.non.separable.initializer");
                        }
                    }
                }
            }
        }
    }
        // where
        /** check if a class is a subtype of Serializable, if that is available. */
        private boolean isSerializable(ClassSymbol c) {
            try {
                syms.serializableType.complete();
            }
            catch (CompletionFailure e) {
                return false;
            }
            return types.isSubtype(c.type, syms.serializableType);
        }

        /** Check that an appropriate serialVersionUID member is defined. */
        private void checkSerialVersionUID(JCClassDecl tree, ClassSymbol c) {

            // check for presence of serialVersionUID
            Scope.Entry e = c.members().lookup(names.serialVersionUID);
            while (e.scope != null && e.sym.kind != VAR) e = e.next();
            if (e.scope == null) {
                log.warning(tree.pos(), "missing.SVUID", c);
                return;
            }

            // check that it is static final
            VarSymbol svuid = (VarSymbol)e.sym;
            if ((svuid.flags() & (STATIC | FINAL)) !=
                (STATIC | FINAL))
                log.warning(TreeInfo.diagnosticPositionFor(svuid, tree), "improper.SVUID", c);

            // check that it is long
            else if (svuid.type.tag != TypeTags.LONG)
                log.warning(TreeInfo.diagnosticPositionFor(svuid, tree), "long.SVUID", c);

            // check constant
            else if (svuid.getConstValue() == null)
                log.warning(TreeInfo.diagnosticPositionFor(svuid, tree), "constant.SVUID", c);
        }

    private Type capture(Type type) {
        return types.capture(type);
    }
}
